from __future__ import annotations

from typing import Any, Iterator

import copy

import numpy as np
from ConfigSpace import Configuration

from smac.acquisition.function.abstract_acquisition_function import (
    AbstractAcquisitionFunction,
)
from smac.acquisition.maximizer.abstract_acquisition_maximizer import (
    AbstractAcquisitionMaximizer,
)
from smac.callback.callback import Callback
from smac.initial_design import AbstractInitialDesign
from smac.model.abstract_model import AbstractModel
from smac.random_design.abstract_random_design import AbstractRandomDesign
from smac.runhistory.encoder.abstract_encoder import AbstractRunHistoryEncoder
from smac.runhistory.runhistory import RunHistory
from smac.scenario import Scenario
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


logger = get_logger(__name__)


class ConfigSelector:
    """The config selector handles the surrogate model and the acquisition function. Based on both components, the next
    configuration is selected.

    Parameters
    ----------
    retrain_after : int, defaults to 8
        How many configurations should be returned before the surrogate model is retrained.
    retries : int, defaults to 8
        How often to retry receiving a new configuration before giving up.
    min_trials: int, defaults to 1
        How many samples are required to train the surrogate model. If budgets are involved,
        the highest budgets are checked first. For example, if min_trials is three, but we find only
        two trials in the runhistory for the highest budget, we will use trials of a lower budget
        instead.
    """

    def __init__(
        self,
        scenario: Scenario,
        *,
        retrain_after: int = 8,
        retries: int = 16,
        min_trials: int = 1,
    ) -> None:
        # Those are the configs sampled from the passed initial design
        # Selecting configurations from initial design
        self._initial_design_configs: list[Configuration] = []

        # Set classes globally
        self._scenario = scenario
        self._runhistory: RunHistory | None = None
        self._runhistory_encoder: AbstractRunHistoryEncoder | None = None
        self._model: AbstractModel | None = None
        self._acquisition_maximizer: AbstractAcquisitionMaximizer | None = None
        self._acquisition_function: AbstractAcquisitionFunction | None = None
        self._random_design: AbstractRandomDesign | None = None
        self._callbacks: list[Callback] = []

        # And other variables
        self._retrain_after = retrain_after
        self._previous_entries = -1
        self._predict_x_best = True
        self._min_trials = min_trials
        self._considered_budgets: list[float | int | None] = [None]

        # How often to retry receiving a new configuration
        # (counter increases if the received config was already returned before)
        self._retries = retries

        # Processed configurations should be stored here; this is important to not return the same configuration twice
        self._processed_configs: list[Configuration] = []

    def _set_components(
        self,
        initial_design: AbstractInitialDesign,
        runhistory: RunHistory,
        runhistory_encoder: AbstractRunHistoryEncoder,
        model: AbstractModel,
        acquisition_maximizer: AbstractAcquisitionMaximizer,
        acquisition_function: AbstractAcquisitionFunction,
        random_design: AbstractRandomDesign,
        callbacks: list[Callback] = None,
    ) -> None:
        self._runhistory = runhistory
        self._runhistory_encoder = runhistory_encoder
        self._model = model
        self._acquisition_maximizer = acquisition_maximizer
        self._acquisition_function = acquisition_function
        self._random_design = random_design
        self._callbacks = callbacks if callbacks is not None else []

        self._initial_design_configs = initial_design.select_configurations()
        if len(self._initial_design_configs) == 0:
            raise RuntimeError("SMAC needs initial configurations to work.")

    @property
    def meta(self) -> dict[str, Any]:
        """Returns the meta data of the created object."""
        return {
            "name": self.__class__.__name__,
            "retrain_after": self._retrain_after,
            "retries": self._retries,
            "min_trials": self._min_trials,
        }

    def __iter__(self) -> Iterator[Configuration]:
        """This method returns the next configuration to evaluate. It ignores already processed configurations, i.e.,
        the configurations from the runhistory, if the runhistory is not empty.
        The method (after yielding the initial design configurations) trains the surrogate model, maximizes the
        acquisition function and yields ``n`` configurations. After the ``n`` configurations, the surrogate model is
        trained again, etc. The program stops if ``retries`` was reached within each iteration. A configuration
        is ignored, if it was used already before.

        Note
        ----
        When SMAC continues a run, processed configurations from the runhistory are ignored. For example, if the
        intitial design configurations already have been processed, they are ignored here. After the run is
        continued, however, the surrogate model is trained based on the runhistory in all cases.

        Returns
        -------
        next_config : Iterator[Configuration]
            The next configuration to evaluate.
        """
        assert self._runhistory is not None
        assert self._runhistory_encoder is not None
        assert self._model is not None
        assert self._acquisition_maximizer is not None
        assert self._acquisition_function is not None
        assert self._random_design is not None

        self._processed_configs = self._runhistory.get_configs()

        # We add more retries because there could be a case in which the processed configs are sampled again
        self._retries += len(self._processed_configs)

        logger.debug("Search for the next configuration...")
        self._call_callbacks_on_start()

        # First: We return the initial configurations
        for config in self._initial_design_configs:
            if config not in self._processed_configs:
                self._processed_configs.append(config)
                self._call_callbacks_on_end(config)
                yield config
                self._call_callbacks_on_start()

        # We want to generate configurations endlessly
        while True:
            # Cost value of incumbent configuration (required for acquisition function).
            # If not given, it will be inferred from runhistory or predicted.
            # If not given and runhistory is empty, it will raise a ValueError.
            incumbent_value: float | None = None

            # Everytime we re-train the surrogate model, we also update our multi-objective algorithm
            if (mo := self._runhistory_encoder.multi_objective_algorithm) is not None:
                mo.update_on_iteration_start()

            X, Y, X_configurations = self._collect_data()
            previous_configs = self._runhistory.get_configs()

            if X.shape[0] == 0:
                # Only return a single point to avoid an overly high number of random search iterations.
                # We got rid of random search here and replaced it with a simple configuration sampling from
                # the configspace.
                logger.debug("No data available to train the model. Sample a random configuration.")

                config = self._scenario.configspace.sample_configuration()
                self._call_callbacks_on_end(config)
                yield config
                self._call_callbacks_on_start()

                # Important to continue here because we still don't have data available
                continue

            # Check if X/Y differs from the last run, otherwise use cached results
            if self._previous_entries != Y.shape[0]:
                self._model.train(X, Y)

                x_best_array: np.ndarray | None = None
                if incumbent_value is not None:
                    best_observation = incumbent_value
                else:
                    if self._runhistory.empty():
                        raise ValueError("Runhistory is empty and the cost value of the incumbent is unknown.")

                    x_best_array, best_observation = self._get_x_best(X_configurations)

                self._acquisition_function.update(
                    model=self._model,
                    eta=best_observation,
                    incumbent_array=x_best_array,
                    num_data=len(self._get_evaluated_configs()),
                    X=X_configurations,
                )

            # We want to cache how many entries we used because if we have the same number of entries
            # we don't need to train the next time
            self._previous_entries = Y.shape[0]

            # Now we maximize the acquisition function
            challengers = self._acquisition_maximizer.maximize(
                previous_configs,
                random_design=self._random_design,
            )

            counter = 0
            failed_counter = 0
            for config in challengers:
                if config not in self._processed_configs:
                    counter += 1
                    self._processed_configs.append(config)
                    self._call_callbacks_on_end(config)
                    yield config
                    retrain = counter == self._retrain_after
                    self._call_callbacks_on_start()

                    # We break to enforce a new iteration of the while loop (i.e. we retrain the surrogate model)
                    if retrain:
                        logger.debug(
                            f"Yielded {counter} configurations. Start new iteration and retrain surrogate model."
                        )
                        break
                else:
                    failed_counter += 1

                    # We exit the loop if we have tried to add the same configuration too often
                    if failed_counter == self._retries:
                        logger.warning(f"Could not return a new configuration after {self._retries} retries." "")
                        return

    def _call_callbacks_on_start(self) -> None:
        for callback in self._callbacks:
            callback.on_next_configurations_start(self)

    def _call_callbacks_on_end(self, config: Configuration) -> None:
        """Calls ``on_next_configurations_end`` of the registered callbacks."""
        # For safety reasons: Return a copy of the config
        if len(self._callbacks) > 0:
            config = copy.deepcopy(config)

        for callback in self._callbacks:
            callback.on_next_configurations_end(self, config)

    def _collect_data(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Collects the data from the runhistory to train the surrogate model. In the case of budgets, the data
        collection strategy is as follows: Looking from highest to lowest budget, return those observations
        that support at least ``self._min_trials`` points.

        If no budgets are used, this is equivalent to returning all observations.
        """
        assert self._runhistory is not None
        assert self._runhistory_encoder is not None

        # If we use a float value as a budget, we want to train the model only on the highest budget
        unique_budgets: set[float] = {run_key.budget for run_key in self._runhistory if run_key.budget is not None}

        available_budgets: list[float] | list[None]
        if len(unique_budgets) > 0:
            # Sort available budgets from highest to lowest budget
            available_budgets = sorted(unique_budgets, reverse=True)
        else:
            available_budgets = [None]

        # Get #points per budget and if there are enough samples, then build a model
        for b in available_budgets:
            X, Y = self._runhistory_encoder.transform(budget_subset=[b])

            if X.shape[0] >= self._min_trials:
                self._considered_budgets = [b]

                # Possible add running configs?
                configs_array = self._runhistory_encoder.get_configurations(budget_subset=self._considered_budgets)

                return X, Y, configs_array

        return (
            np.empty(shape=[0, 0]),
            np.empty(
                shape=[
                    0,
                ]
            ),
            np.empty(shape=[0, 0]),
        )

    def _get_evaluated_configs(self) -> list[Configuration]:
        assert self._runhistory is not None
        return self._runhistory.get_configs_per_budget(budget_subset=self._considered_budgets)

    def _get_x_best(self, X: np.ndarray) -> tuple[np.ndarray, float]:
        """Get value, configuration, and array representation of the *best* configuration.

        The definition of best varies depending on the argument ``predict``. If set to `True`,
        this function will return the stats of the best configuration as predicted by the model,
        otherwise it will return the stats for the best observed configuration.

        Parameters
        ----------
        predict : bool
            Whether to use the predicted or observed best.

        Returns
        -------
        float
        np.ndarry
        Configuration
        """
        if self._predict_x_best:
            model = self._model
            costs = list(
                map(
                    lambda x: (
                        model.predict_marginalized(x.reshape((1, -1)))[0][0][0],  # type: ignore
                        x,
                    ),
                    X,
                )
            )
            costs = sorted(costs, key=lambda t: t[0])
            x_best_array = costs[0][1]
            best_observation = costs[0][0]

        # else:
        #    all_configs = self._runhistory.get_configs_per_budget(budget_subset=self._considered_budgets)
        #    x_best = self._incumbent
        #    x_best_array = convert_configurations_to_array(all_configs)
        #    best_observation = self._runhistory.get_cost(x_best)
        #    best_observation_as_array = np.array(best_observation).reshape((1, 1))

        #    # It's unclear how to do this for inv scaling and potential future scaling.
        #    # This line should be changed if necessary
        #    best_observation = self._runhistory_encoder.transform_response_values(best_observation_as_array)
        #    best_observation = best_observation[0][0]

        return x_best_array, best_observation



from __future__ import annotations

from abc import abstractmethod

from ConfigSpace import ConfigurationSpace
from src.datasets.dataset import Dataset


class Model:
    def __init__(self, dataset: Dataset | None):
        self._dataset = dataset

    @property
    def dataset(self) -> Dataset | None:
        return self._dataset

    @property
    @abstractmethod
    def configspace(self) -> ConfigurationSpace:
        raise NotImplementedError


class SingleObjectiveModel(Model):
    @abstractmethod
    def train(self) -> float:
        raise NotImplementedError


class MultiObjectiveModel(Model):
    @abstractmethod
    def train(self) -> list[float]:
        raise NotImplementedError



from smac.model.random_forest.abstract_random_forest import AbstractRandomForest
from smac.model.random_forest.random_forest import RandomForest

__all__ = [
    "AbstractRandomForest",
    "RandomForest",
]



from __future__ import annotations

from typing import Any

import warnings

from ConfigSpace.configuration_space import Configuration
from ConfigSpace.hyperparameters import Constant
from scipy.stats.qmc import Sobol

from smac.initial_design.abstract_initial_design import AbstractInitialDesign
from smac.utils.configspace import transform_continuous_designs

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class SobolInitialDesign(AbstractInitialDesign):
    """Sobol sequence design with a scrambled Sobol sequence. See
    https://scipy.github.io/devdocs/reference/generated/scipy.stats.qmc.Sobol.html for further information.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        if len(list(self._configspace.values())) > 21201:
            raise ValueError(
                "The default initial design Sobol sequence can only handle up to 21201 dimensions. "
                "Please use a different initial design, such as the Latin Hypercube design."
            )

    def _select_configurations(self) -> list[Configuration]:
        params = list(self._configspace.values())

        constants = 0
        for p in params:
            if isinstance(p, Constant):
                constants += 1

        dim = len(params) - constants
        sobol_gen = Sobol(d=dim, scramble=True, seed=self._rng.randint(low=0, high=10000000))

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            sobol = sobol_gen.random(self._n_configs)

        return transform_continuous_designs(design=sobol, origin="Initial Design: Sobol", configspace=self._configspace)



from smac.multi_objective.abstract_multi_objective_algorithm import (
    AbstractMultiObjectiveAlgorithm,
)
from smac.multi_objective.aggregation_strategy import MeanAggregationStrategy
from smac.multi_objective.parego import ParEGO

__all__ = [
    "AbstractMultiObjectiveAlgorithm",
    "MeanAggregationStrategy",
    "ParEGO",
]



from smac.runhistory.dataclasses import (
    InstanceSeedBudgetKey,
    InstanceSeedKey,
    TrialInfo,
    TrialKey,
    TrialValue,
)
from smac.runhistory.enumerations import StatusType
from smac.runhistory.runhistory import RunHistory

__all__ = [
    "RunHistory",
    "TrialKey",
    "InstanceSeedBudgetKey",
    "InstanceSeedKey",
    "TrialValue",
    "TrialInfo",
    "StatusType",
]



import numpy as np
from src.datasets.breast_cancer import BreastCancerDataset
from src.datasets.digits import DigitsDataset
from src.models.branin import Branin
from src.models.ac_branin import ACBranin
from src.datasets.ac_branin import ACBraninDataset
from src.models.himmelblau import HimmelblauModel
from src.models.mlp import MLPModel
from src.models.svm import SVMModel
from src.tasks.task import Task

TASKS = [
    ###### Branin ########
    Task(
        name="Branin (Blackbox)",
        model=Branin(),
        deterministic=True,
        objectives=["regret"],
        optimization_type="bb",
        n_trials=200,
        walltime_limit=60,
        max_config_calls=1,
    ),
    Task(
        name="Branin (HPO)",
        model=Branin(),
        deterministic=True,
        objectives=["regret"],
        optimization_type="hpo",
        n_trials=200,
        walltime_limit=60,
        max_config_calls=1,
    ),
    Task(
        name="Branin (HPO, 8 workers)",
        model=Branin(),
        deterministic=True,
        objectives=["regret"],
        optimization_type="hpo",
        n_trials=200,
        walltime_limit=60,
        max_config_calls=1,
        n_workers=8,
    ),
    Task(
        name="Branin (AC)",
        model=ACBranin(ACBraninDataset()),
        deterministic=True,
        objectives=["regret"],
        optimization_type="ac",
        n_trials=300,
        walltime_limit=1720,
        max_config_calls=10000,
        use_instances=True,
    ),
    ######## Himmelblau ########
    Task(
        name="Himmelblau (Blackbox)",
        model=HimmelblauModel(),
        deterministic=True,
        objectives=["cost"],
        optimization_type="bb",
        n_trials=300,
        walltime_limit=60,
        max_config_calls=1,
    ),
    Task(
        name="Himmelblau (HPO)",
        model=HimmelblauModel(),
        deterministic=True,
        objectives=["cost"],
        optimization_type="hpo",
        n_trials=300,
        walltime_limit=60,
        max_config_calls=1,
    ),
    Task(
        name="Himmelblau (HPO, 8 workers)",
        model=HimmelblauModel(),
        deterministic=True,
        objectives=["cost"],
        optimization_type="hpo",
        n_trials=300,
        walltime_limit=60,
        max_config_calls=1,
        n_workers=8,
    ),
    ######## SVM-Digits (HB) ########
    Task(
        name="SVM-Digits",
        model=SVMModel(DigitsDataset()),
        deterministic=False,
        objectives=["1-accuracy"],
        optimization_type="hpo",
        n_trials=500,
        walltime_limit=np.inf,
        use_instances=True,
        max_config_calls=50,
    ),
    Task(
        name="SVM-Digits (HB)",
        model=SVMModel(DigitsDataset()),
        deterministic=False,
        objectives=["1-accuracy"],
        optimization_type="mf",
        n_trials=500,
        walltime_limit=np.inf,
        use_instances=True,
        min_budget=10,
        max_budget=50,
        incumbent_selection="highest_observed_budget",
        n_seeds=2,
    ),
    Task(
        name="SVM-Digits (HB, 8 workers)",
        model=SVMModel(DigitsDataset()),
        deterministic=False,
        objectives=["1-accuracy"],
        optimization_type="mf",
        n_trials=500,
        walltime_limit=np.inf,
        use_instances=True,
        min_budget=10,
        max_budget=50,
        incumbent_selection="highest_observed_budget",
        n_seeds=2,
        n_workers=8,
    ),
    ######## BreastCancer (HB) ########
    Task(
        name="MLP-BreastCancer (HB)",
        model=MLPModel(BreastCancerDataset()),
        deterministic=False,
        objectives=["1-accuracy"],
        optimization_type="mf",
        min_budget=1,
        max_budget=25,
        n_trials=1500,
        walltime_limit=480,
        n_seeds=1,
        incumbent_selection="highest_budget",
    ),
    Task(
        name="MLP-BreastCancer (HB, 8 workers)",
        model=MLPModel(BreastCancerDataset()),
        deterministic=False,
        objectives=["1-accuracy"],
        optimization_type="mf",
        min_budget=1,
        max_budget=25,
        n_trials=1500,
        walltime_limit=480,
        n_seeds=1,
        incumbent_selection="highest_budget",
        n_workers=8,
    ),
    ######## BreastCancer (SH) ########
    Task(
        name="MLP-BreastCancer (SH)",
        model=MLPModel(BreastCancerDataset()),
        deterministic=False,
        objectives=["1-accuracy"],
        optimization_type="mf",
        min_budget=1,
        max_budget=25,
        n_trials=1500,
        walltime_limit=480,
        n_seeds=1,
        incumbent_selection="highest_budget",
        intensifier="successive_halving",
    ),
    Task(
        name="MLP-BreastCancer (SH, 8 workers)",
        model=MLPModel(BreastCancerDataset()),
        deterministic=False,
        objectives=["1-accuracy"],
        optimization_type="mf",
        min_budget=1,
        max_budget=25,
        n_trials=1500,
        walltime_limit=480,
        n_seeds=1,
        incumbent_selection="highest_budget",
        intensifier="successive_halving",
        n_workers=8,
    ),
]



from __future__ import annotations

from typing import Any

import numpy as np
import sklearn.gaussian_process.kernels as kernels

from smac.model.gaussian_process.kernels.base_kernels import AbstractKernel
from smac.model.gaussian_process.priors.abstract_prior import AbstractPrior

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class HammingKernel(
    AbstractKernel,
    kernels.StationaryKernelMixin,
    kernels.NormalizedKernelMixin,
    kernels.Kernel,
):
    """Hamming kernel implementation."""

    def __init__(
        self,
        length_scale: float | tuple[float, ...] | np.ndarray = 1.0,
        length_scale_bounds: tuple[float, float] | list[tuple[float, float]] | np.ndarray = (1e-5, 1e5),
        operate_on: np.ndarray | None = None,
        has_conditions: bool = False,
        prior: AbstractPrior | None = None,
    ) -> None:
        self.length_scale = length_scale
        self.length_scale_bounds = length_scale_bounds

        super().__init__(
            operate_on=operate_on,
            has_conditions=has_conditions,
            prior=prior,
        )

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta

        length_scale = self.length_scale
        if isinstance(length_scale, np.ndarray):
            length_scale = length_scale.tolist()

        length_scale_bounds = self.length_scale_bounds
        if isinstance(length_scale_bounds, np.ndarray):
            length_scale_bounds = length_scale_bounds.tolist()

        meta.update(
            {
                "length_scale": length_scale,
                "lengthscale_bounds": length_scale_bounds,
            }
        )

        return meta

    @property
    def hyperparameter_length_scale(self) -> kernels.Hyperparameter:
        """Hyperparameter of the length scale."""
        length_scale = self.length_scale
        anisotropic = np.iterable(length_scale) and len(length_scale) > 1  # type: ignore

        if anisotropic:
            return kernels.Hyperparameter(
                "length_scale",
                "numeric",
                self.length_scale_bounds,
                len(length_scale),  # type: ignore
            )

        return kernels.Hyperparameter(
            "length_scale",
            "numeric",
            self.length_scale_bounds,
        )

    def _call(
        self,
        X: np.ndarray,
        Y: np.ndarray | None = None,
        eval_gradient: bool = False,
        active: np.ndarray | None = None,
    ) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        X = np.atleast_2d(X)
        length_scale = kernels._check_length_scale(X, self.length_scale)

        if Y is None:
            Y = X
        elif eval_gradient:
            raise ValueError("gradient can be evaluated only when Y != X")
        else:
            Y = np.atleast_2d(Y)

        indicator = np.expand_dims(X, axis=1) != Y
        K = (-1 / (2 * length_scale**2) * indicator).sum(axis=2)
        K = np.exp(K)

        if active is not None:
            K = K * active

        if eval_gradient:
            # dK / d theta = (dK / dl) * (dl / d theta)
            # theta = log(l) => dl / d (theta) = e^theta = l
            # dK / d theta = l * dK / dl

            # dK / dL computation
            if np.iterable(length_scale) and length_scale.shape[0] > 1:  # type: ignore
                grad = np.expand_dims(K, axis=-1) * np.array(indicator, dtype=np.float32)
            else:
                grad = np.expand_dims(K * np.sum(indicator, axis=2), axis=-1)

            grad *= 1 / length_scale**3

            return K, grad

        return K



from smac.runner.abstract_runner import AbstractRunner
from smac.runner.dask_runner import DaskParallelRunner
from smac.runner.exceptions import (
    FirstRunCrashedException,
    TargetAlgorithmAbortException,
)
from smac.runner.target_function_runner import TargetFunctionRunner

__all__ = [
    # Runner
    "AbstractRunner",
    "TargetFunctionRunner",
    "DaskParallelRunner",
    # Exceptions
    "TargetAlgorithmAbortException",
    "FirstRunCrashedException",
]



from __future__ import annotations

from typing import Any

import numpy as np

from smac import constants
from smac.runhistory.encoder.encoder import RunHistoryEncoder
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


logger = get_logger(__name__)


class RunHistorySqrtScaledEncoder(RunHistoryEncoder):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        if self._instances is not None and len(self._instances) > 1:
            raise NotImplementedError("Handling more than one instance is not supported for sqrt scaled cost.")

    def transform_response_values(self, values: np.ndarray) -> np.ndarray:
        """Transform the response values by linearly scaling them between zero and one and then using the
        square root.
        """
        # Subtract the difference between the percentile and the minimum
        min_y = self._min_y - (self._percentile - self._min_y)

        # Minimal value to avoid numerical issues in the log scaling below
        min_y -= constants.VERY_SMALL_NUMBER

        # Linear scaling: prevent diving by zero
        min_y[np.where(min_y == self._max_y)] *= 1 - 10**-10

        values = (values - min_y) / (self._max_y - min_y)
        values = np.sqrt(values)

        return values



"""Constants used in SMAC, e.g. maximum number of cutoffs, very small number, etc."""
__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


MAXINT = 2**31 - 1
MINIMAL_COST_FOR_LOG = 0.00001
MAX_CUTOFF = 65535
VERY_SMALL_NUMBER = 1e-10  # The very small number is used to avoid numerical instabilities or dividing by zero
N_TREES = 10



# from __future__ import annotations

# from typing import Any, Dict, Tuple

# import copy
# import math

# import gpytorch
# import numpy as np
# import torch
# from gpytorch import settings
# from gpytorch.kernels import Kernel, MaternKernel, ProductKernel, ScaleKernel
# from gpytorch.lazy import (
#     DiagLazyTensor,
#     MatmulLazyTensor,
#     PsdSumLazyTensor,
#     RootLazyTensor,
#     delazify,
# )
# from gpytorch.likelihoods import GaussianLikelihood
# from gpytorch.means.mean import Mean
# from gpytorch.utils.cholesky import psd_safe_cholesky
# from sklearn.gaussian_process.kernels import Kernel as SKLKernels

# from smac.model.gaussian_process.kernels import ConstantKernel, WhiteKernel


# class MixedKernel(ProductKernel):
#     """
#     A special form of ProductKernel. It is composed of a cont_kernel and a cat_kernel that work with continuous and
#     categorical parameters, respectively. Its forward pass allows an additional parameter to determine if only
#     cont_kernel is applied to the input.
#     """

#     def __init__(self, cont_kernel: Kernel, cat_kernel: Kernel):
#         kernels = cont_kernel.kernels if isinstance(cont_kernel, ProductKernel) else [cont_kernel]
#         kernels += cat_kernel.kernels if isinstance(cat_kernel, ProductKernel) else [cat_kernel]
#         super().__init__(*kernels)
#         self.cont_kernel = cont_kernel
#         self.cat_kernel = cat_kernel

#     def forward(
#         self, x1: torch.Tensor, x2: torch.Tensor, diag: bool = False, cont_only: bool = False, **params: Any
#     ) -> gpytorch.lazy.LazyTensor:
#         """Compute kernel values, if cont_only is True, then the categorical kernel is omitted"""
#         if not cont_only:
#             return super().forward(x1, x2, diag, **params)
#         else:
#             return self.cont_kernel(x1, x2, diag, **params)


# def construct_gp_kernel(
#     kernel_kwargs: Dict[str, Any], cont_dims: np.ndarray, cat_dims: np.ndarray
# ) -> Kernel | SKLKernels:
#     """
#     Construct a GP kernel with the given kernel init argument, the cont_dims, and cat_dims of the problem. Since the
#     subspace might not have the same number of dimensions as the global search space.
#     We need to reconstruct the kernel every time when a new subspace is generated.

#     Parameters
#     ----------
#     kernel_kwargs: Dict[str, Any]
#         kernel kwargs. Arguments to initialize the kernels. It needs to contain the following items:
#             cont_kernel: type of continuous kernels
#             cont_kernel_kwargs: additional arguments for continuous kernels, for instance, length constraints and
# prior
#             cat_kernel: type of categorical kernels
#             cat_kernel_kwargs: additional arguments for categorical kernels, for instance, length constraints and
# prior
#             scale_kernel: type of scale kernels
#             scale_kernel_kwargs: additional arguments for scale kernels,  for instance, length constraints and prior
#     cont_dims: np.ndarray
#         dimensions of continuous hyperparameters
#     cat_dims: np.ndarray
#         dimensions of categorical hyperparameters
#     Returns
#     -------
#     kernel: Kernel | SKLKernels
#         constructed kernels

#     """
#     if len(cont_dims) > 0:
#         cont_kernel_class = kernel_kwargs.get("cont_kernel", MaternKernel)
#         cont_kernel_kwargs = kernel_kwargs.get("cont_kernel_kwargs", {})
#         cont_kernel = cont_kernel_class(
#             ard_num_dims=cont_dims.shape[-1], active_dims=tuple(cont_dims), **cont_kernel_kwargs
#         ).double()

#     if len(cat_dims) > 0:
#         cat_kernel_class = kernel_kwargs.get("cat_kernel", MaternKernel)
#         cat_kernel_kwargs = kernel_kwargs.get("cat_kernel_kwargs", {})
#         cat_kernel = cat_kernel_class(
#             ard_num_dims=cat_dims.shape[-1], active_dims=tuple(cat_dims), **cat_kernel_kwargs
#         ).double()

#     if len(cont_dims) > 0 and len(cat_dims) > 0:
#         if isinstance(cont_kernel, SKLKernels):
#             base_kernel = cont_kernel * cat_kernel
#         else:
#             base_kernel = MixedKernel(cont_kernel=cont_kernel, cat_kernel=cat_kernel)
#     elif len(cont_dims) > 0 and len(cat_dims) == 0:
#         base_kernel = cont_kernel
#     elif len(cont_dims) == 0 and len(cat_dims) > 0:
#         base_kernel = cat_kernel
#     else:
#         raise ValueError("Either cont_dims or cat_dims must exist!")

#     if isinstance(base_kernel, SKLKernels):
#         scale_kernel_class = kernel_kwargs.get("scale_kernel", ConstantKernel)
#         scale_kernel_kwargs = kernel_kwargs.get("scale_kernel_kwargs", {})
#         scale_kernel = scale_kernel_class(**scale_kernel_kwargs)

#         noise_kernel_class = kernel_kwargs.get("noise_kernel", WhiteKernel)
#         noise_kernel_kwargs = kernel_kwargs.get("noise_kernel_kwargs", {})
#         noise_kernel = noise_kernel_class(**noise_kernel_kwargs)

#         gp_kernel = scale_kernel * base_kernel + noise_kernel
#     else:
#         scale_kernel_class = kernel_kwargs.get("scale_kernel", ScaleKernel)
#         scale_kernel_kwargs = kernel_kwargs.get("scale_kernel_kwargs", {})
#         gp_kernel = scale_kernel_class(base_kernel=base_kernel, **scale_kernel_kwargs)

#     return gp_kernel


# class FITCKernel(Kernel):
#     def __init__(
#         self,
#         base_kernel: Kernel,
#         X_inducing: torch.Tensor,
#         likelihood: GaussianLikelihood,
#         X_out: torch.Tensor,
#         y_out: torch.Tensor,
#         active_dims: Tuple[int] | None = None,
#     ):
#         r"""A reimplementation of FITC Kernel that computes the posterior explicitly for globally augmented local GP.
#         This should work exactly the same as a gpytorch.kernel.InducingPointKernel.
#          However, it takes much less time when combined with LGPGA.
#          References: Edward Snelson and Zoubin Ghahramani. Sparse Gaussian processes using pseudo-inputs. Advances in
#          Neural Information Processing Systems 18, Cambridge, Massachusetts, 2006. The MIT Press.
#          https://papers.nips.cc/paper/2005/hash/4491777b1aa8b5b32c2e8666dbe1a495-Abstract.html

#         Mean value is computed with:
#         \mathbf{\mu_{l'}}  = \mathbf{K_{l',u} \Sigma K_{u,1} \Lambda}^{-1}\mathbf{y_g} \label{eq:mean_sgp}
#         and variance value:
#         \mathbf{\sigma}^2_{l'} = \mathbf{K_{l',l'}} - \mathbf{Q_{l', l'} + \mathbf{K_{l', u}\Sigma K_{u, l'}}}
#         \mathbf{\Sigma} = (\mathbf{K_{u,u}} + \mathbf{K_{u, g} \Lambda}^{-1}\mathbf{K_{g,u}})^{-1}
#         \mathbf{\Lambda} = diag[\mathbf{K_{g,g}-Q_{g,g}} + \sigma^2_{noise}\idenmat]
#         ----------
#         base_kernel: Kernel
#             base kernel function
#         X_inducing: torch.Tensor (N_inducing, D)
#             inducing points, a torch tensor with shape (N_inducing, D), N_inducing is the number of the inducing
# points
#         likelihood: GaussianLikelihood
#             GP likelihood
#         X_out: torch.Tensor (N_out,D)
#             data features outside the subregion, it needs to be of size (N_out, D), N_out is the number of points
#             outside the subspace
#         y_out: torch.Tensor
#             data observations outside the subregion
#         active_dims: Tuple[int] | None = None
#             Set this if you want to compute the covariance of only a few input dimensions. The ints
#             corresponds to the indices of the dimensions. Default: `None`.
#         """
#         super(FITCKernel, self).__init__(active_dims=active_dims)
#         self.has_lengthscale = base_kernel.has_lengthscale
#         self.base_kernel = base_kernel
#         self.likelihood = likelihood

#         if X_inducing.ndimension() == 1:
#             X_inducing = X_inducing.unsqueeze(-1)

#         self.X_out = X_out
#         self.y_out = y_out
#         self.register_parameter(name="X_inducing", parameter=torch.nn.Parameter(X_inducing))

#     def _clear_cache(self) -> None:
#         if hasattr(self, "_cached_kernel_mat"):
#             del self._cached_kernel_mat
#         if hasattr(self, "_cached_inducing_sigma"):
#             del self._cached_inducing_sigma
#         if hasattr(self, "_cached_poster_mean_mat"):
#             del self._cached_poster_mean_mat
#         if hasattr(self, "_train_cached_k_u1"):
#             del self._train_cached_k_u1
#         if hasattr(self, "_train_cached_lambda_diag_inv"):
#             del self._train_cached_lambda_diag_inv
#         if hasattr(self, "_train_cached_posterior_mean"):
#             del self._train_cached_posterior_mean
#         if hasattr(self, "_cached_kernel_inv_root"):
#             del self._cached_kernel_inv_root

#     @property
#     def _inducing_mat(self) -> torch.Tensor:
#         """
#         Computes inducing matrix, K(X_inducing, X_inducing)

#         Returns
#         -------
#         res: torch.Tensor (N_inducing, N_inducing)
#             K(X_inducing, X_inducing)
#         """
#         if not self.training and hasattr(self, "_cached_kernel_mat"):
#             return self._cached_kernel_mat
#         else:
#             res = delazify(self.base_kernel(self.X_inducing, self.X_inducing))
#             if not self.training:
#                 self._cached_kernel_mat = res  # type: torch.Tensor
#             return res

#     @property
#     def _inducing_inv_root(self) -> torch.Tensor:
#         """
#         Computes the inverse of the inducing matrix: K_inv(X_inducing, X_inducing) = K(X_inducing, X_inducing)^(-1)

#         Returns
#         -------
#         res: torch.Tensor (N_inducing, N_inducing)
#             K_inv(X_inducing, X_inducing)
#         """
#         if not self.training and hasattr(self, "_cached_kernel_inv_root"):
#             return self._cached_kernel_inv_root
#         else:
#             chol = psd_safe_cholesky(self._inducing_mat, upper=True, jitter=settings.cholesky_jitter.value())
#             eye = torch.eye(chol.size(-1), device=chol.device, dtype=chol.dtype)
#             inv_root = torch.triangular_solve(eye, chol)[0]

#             res = inv_root
#             if not self.training:
#                 self._cached_kernel_inv_root = res  # type: torch.Tensor
#             return res

#     @property
#     def _k_u1(self) -> torch.Tensor:
#         """
#         Computes the covariance matrix between the X_inducing and X_out : K(X_inducing, X_out)

#         Returns
#         -------
#         res: torch.Tensor (N_inducing, N_out)
#             K(X_inducing, X_out)
#         """
#         if not self.training and hasattr(self, "_cached_k_u1"):
#             return self._cached_k_u1
#         else:
#             res = delazify(self.base_kernel(self.X_inducing, self.X_out))
#             if not self.training:
#                 self._cached_k_u1 = res  # type: torch.Tensor
#             else:
#                 self._train_cached_k_u1 = res  # type: torch.Tensor
#             return res

#     @property
#     def _lambda_diag_inv(self) -> torch.Tensor:
#         r"""Computes the inverse of lambda matrix, it is computed by
#         \Lambda = diag[\mathbf{K_{X_out,X_out}-Q_{X_out,X_out}} + \sigma^2_{noise}\idenmat] and
#         Q{X_out, X_out} = K(X_out, X_inducing) K^{-1}(X_inducing,X_inducing) K(X_inducing, X_out)

#         Returns
#         -------
#         res: torch.Tensor (N_out, N_out)
#             inverse of the diagonal matrix lambda
#         """
#         if not self.training and hasattr(self, "_cached_lambda_diag_inv"):
#             return self._cached_lambda_diag_inv
#         else:
#             diag_k11 = delazify(self.base_kernel(self.X_out, diag=True))

#             diag_q11 = delazify(RootLazyTensor(self._k_u1.transpose(-1, -2).matmul(self._inducing_inv_root))).diag()

#             # Diagonal correction for predictive posterior
#             correction = (diag_k11 - diag_q11).clamp(0, math.inf)

#             sigma = self.likelihood._shaped_noise_covar(correction.shape).diag()

#             res = delazify(DiagLazyTensor((correction + sigma).reciprocal()))

#             if not self.training:
#                 self._cached_lambda_diag_inv = res  # type: torch.Tensor
#             else:
#                 self._train_cached_lambda_diag_inv = res  # type: torch.Tensor
#             return res

#     @property
#     def _inducing_sigma(self) -> torch.Tensor:
#         r"""Computes the inverse of lambda matrix, it is computed by
#         \mathbf{\Sigma} = (\mathbf{K_{X_inducing,X_inducing}} +
#          \mathbf{K_{X_inducing, X_out} \Lambda}^{-1}\mathbf{K_{X_out,X_inducing}})

#         Returns
#         -------
#         res: torch.Tensor (N_inducing, N_inducing)
#             \Sigma
#         """
#         if not self.training and hasattr(self, "_cached_inducing_sigma"):
#             return self._cached_inducing_sigma
#         else:
#             k_u1 = self._k_u1
#             res = PsdSumLazyTensor(
#                 self._inducing_mat,
#                 MatmulLazyTensor(k_u1, MatmulLazyTensor(self._lambda_diag_inv, k_u1.transpose(-1, -2))),
#             )
#             res = delazify(res)
#             if not self.training:
#                 self._cached_inducing_sigma = res  # type: torch.Tensor

#             return res

#     @property
#     def _inducing_sigma_inv_root(self) -> torch.Tensor:
#         r"""Inverse of Sigma matrix:

#         Returns
#         -------
#         res: torch.Tensor (N_inducing, N_inducing)
#             \Sigma ^{-1}
#         """
#         if not self.training and hasattr(self, "_cached_inducing_sigma_inv_root"):
#             return self._cached_inducing_sigma_inv_root
#         else:
#             chol = psd_safe_cholesky(self._inducing_sigma, upper=True, jitter=settings.cholesky_jitter.value())

#             eye = torch.eye(chol.size(-1), device=chol.device, dtype=chol.dtype)
#             inv_root = torch.triangular_solve(eye, chol)[0]
#             res = inv_root
#             if not self.training:
#                 self._cached_inducing_sigma_inv_root = res  # type: torch.Tensor
#             return res

#     @property
#     def _poster_mean_mat(self) -> torch.Tensor:
#         r"""A cached value for computing the posterior mean of a sparse kernel it is defined by
#         \Sigma K_{u, 1} \Lambda}^{-1}\mathbf{y_out}

#         Returns
#         -------
#         res: torch.Tensor (N_inducing, 1)
#             cached posterior mean
#         """
#         if not self.training and hasattr(self, "_cached_poster_mean_mat"):
#             return self._cached_poster_mean_mat
#         else:
#             inducing_sigma_inv_root = self._inducing_sigma_inv_root
#             sigma = RootLazyTensor(inducing_sigma_inv_root)

#             k_u1 = self._k_u1
#             lambda_diag_inv = self._lambda_diag_inv

#             res_mat = delazify(MatmulLazyTensor(sigma, MatmulLazyTensor(k_u1, lambda_diag_inv)))

#             res = torch.matmul(res_mat, self.y_out)

#             if not self.training:
#                 self._cached_poster_mean_mat = res  # type: torch.Tensor
#             return res

#     def _get_covariance(self, x1: torch.Tensor, x2: torch.Tensor) -> gpytorch.lazy.LazyTensor:
#         r"""Compute the posterior covariance matrix of a sparse kernel explicitly

#         Parameters
#         ----------
#         x1: torch.Tensor(N_x1, D)
#             first input of the FITC kernel
#         x2: torch.Tensor(N_x2, D)
#             second input of the FITC kernel

#         Returns
#         -------
#         res: Optional[torch.Tensor (N_x1, 1), PsdSumLazyTensor]
#             a cached value for computing the posterior mean, it
#             is defined by  \Sigma K_{u, 1} \Lambda}^{-1}\mathbf{y_out}
#         """
#         k_x1x2 = self.base_kernel(x1, x2)
#         k_x1u = delazify(self.base_kernel(x1, self.X_inducing))
#         inducing_inv_root = self._inducing_inv_root
#         inducing_sigma_inv_root = self._inducing_sigma_inv_root
#         if torch.equal(x1, x2):
#             q_x1x2 = RootLazyTensor(k_x1u.matmul(inducing_inv_root))

#             s_x1x2 = RootLazyTensor(k_x1u.matmul(inducing_sigma_inv_root))
#         else:
#             k_x2u = delazify(self.base_kernel(x2, self.X_inducing))
#             q_x1x2 = MatmulLazyTensor(
#                 k_x1u.matmul(inducing_inv_root), k_x2u.matmul(inducing_inv_root).transpose(-1, -2)
#             )
#             s_x1x2 = MatmulLazyTensor(
#                 k_x1u.matmul(inducing_sigma_inv_root), k_x2u.matmul(inducing_sigma_inv_root).transpose(-1, -2)
#             )
#         covar = PsdSumLazyTensor(k_x1x2, -1.0 * q_x1x2, s_x1x2)

#         if self.training:
#             k_iu = self.base_kernel(x1, self.X_inducing)
#             sigma = RootLazyTensor(inducing_sigma_inv_root)

#             k_u1 = self._train_cached_k_u1 if hasattr(self, "_train_cached_k_u1") else self._k_u1
#             lambda_diag_inv = (
#                 self._train_cached_lambda_diag_inv
#                 if hasattr(self, "_train_cached_lambda_diag_inv")
#                 else self._lambda_diag_inv
#             )

#             mean = torch.matmul(
#                 delazify(MatmulLazyTensor(k_iu, MatmulLazyTensor(sigma, MatmulLazyTensor(k_u1, lambda_diag_inv)))),
#                 self.y_out,
#             )

#             self._train_cached_posterior_mean = mean  # type: torch.Tensor
#         return covar

#     def posterior_mean(self, inputs: torch.Tensor) -> torch.Tensor:
#         """
#         The posterior mean of the FITC kernel, will serve as the prior mean of the dense kernel.

#         Parameters
#         ----------
#         inputs: torch.Tensor(N_inputs, D)
#             input of the FITC kernel

#         Returns
#         -------
#         res: Torch.Tensor (N_inputs, 1)
#             The posterior mean of the FITC Kernel
#         """
#         if self.training and hasattr(self, "_train_cached_posterior_mean"):
#             return self._train_cached_posterior_mean
#         if inputs.ndimension() == 1:
#             inputs = inputs.unsqueeze(1)

#         k_iu = delazify(self.base_kernel(inputs, self.X_inducing, cont_only=True))
#         poster_mean = self._poster_mean_mat
#         res = torch.matmul(k_iu, poster_mean)
#         return res

#     def forward(
#         self, x1: torch.Tensor, x2: torch.Tensor, diag: bool = False, **kwargs: Any
#     ) -> gpytorch.lazy.LazyTensor:
#         """Compute the kernel function"""
#         covar = self._get_covariance(x1, x2)
#         if self.training:
#             if not torch.equal(x1, x2):
#                 raise RuntimeError("x1 should equal x2 in training mode")

#         if diag:
#             return covar.diag()
#         else:
#             return covar

#     def num_outputs_per_input(self, x1: torch.Tensor, x2: torch.Tensor) -> int:
#         """
#         Number of outputs given the inputs
#         if x1 is of size `n x d` and x2 is size `m x d`, then the size of the kernel
#         will be `(n * num_outputs_per_input) x (m * num_outputs_per_input)`

#         Parameters
#         ----------
#         x1: torch.Tensor
#             the first input of the kernel
#         x2: torch.Tensor
#             the second input of the kernel
#         Returns
#         -------
#         res: int
#             for base kernels such as matern or RBF kernels, this value needs to be 1.
#         """
#         return self.base_kernel.num_outputs_per_input(x1, x2)

#     def __deepcopy__(self, memo: Dict) -> "FITCKernel":
#         replace_inv_root = False
#         replace_kernel_mat = False
#         replace_k_u1 = False
#         replace_lambda_diag_inv = False
#         replace_inducing_sigma = False
#         replace_inducing_sigma_inv_root = False
#         replace_poster_mean = False

#         if hasattr(self, "_cached_kernel_inv_root"):
#             replace_inv_root = True
#             kernel_inv_root = self._cached_kernel_inv_root
#         if hasattr(self, "_cached_kernel_mat"):
#             replace_kernel_mat = True
#             kernel_mat = self._cached_kernel_mat
#         if hasattr(self, "_cached_k_u1"):
#             replace_k_u1 = True
#             k_u1 = self._cached_k_u1
#         if hasattr(self, "_cached_lambda_diag_inv"):
#             replace_lambda_diag_inv = True
#             lambda_diag_inv = self._cached_lambda_diag_inv
#         if hasattr(self, "_cached_inducing_sigma"):
#             replace_inducing_sigma = True
#             inducing_sigma = self._cached_inducing_sigma
#         if hasattr(self, "_cached_inducing_sigma_inv_root"):
#             replace_inducing_sigma_inv_root = True
#             inducing_sigma_inv_root = self._cached_inducing_sigma_inv_root
#         if hasattr(self, "_cached_poster_mean_mat"):
#             replace_poster_mean = True
#             poster_mean_mat = self._cached_poster_mean_mat

#         cp = self.__class__(
#             base_kernel=copy.deepcopy(self.base_kernel),
#             X_inducing=copy.deepcopy(self.X_inducing),
#             X_out=self.X_out,
#             y_out=self.y_out,
#             likelihood=copy.deepcopy(self.likelihood),
#             active_dims=self.active_dims,
#         )

#         if replace_inv_root:
#             cp._cached_kernel_inv_root = kernel_inv_root

#         if replace_kernel_mat:
#             cp._cached_kernel_mat = kernel_mat

#         if replace_k_u1:
#             cp._cached_k_u1 = k_u1

#         if replace_lambda_diag_inv:
#             cp._cached_lambda_diag_inv = lambda_diag_inv

#         if replace_inducing_sigma:
#             cp._cached_inducing_sigma = inducing_sigma

#         if replace_inducing_sigma_inv_root:
#             cp._cached_inducing_sigma_inv_root = inducing_sigma_inv_root

#         if replace_poster_mean:
#             cp._cached_poster_mean_mat = poster_mean_mat

#         return cp


# class FITCMean(Mean):
#     def __init__(self, covar_module: FITCKernel, batch_shape: torch.Size = torch.Size(), **kwargs: Any):
#         """
#         Read the posterior mean value of the given fitc kernel and serve as a prior mean value for the
#         second stage

#         Parameters
#         ----------
#         covar_module: FITCKernel
#             a FITC  kernel
#         batch_shape: torch.size
#             batch size
#         """
#         super(FITCMean, self).__init__()
#         self.covar_module = covar_module
#         self.batch_shape = batch_shape
#         self.covar_module = covar_module

#     def forward(self, input: torch.Tensor) -> torch.Tensor:
#         """
#         Compute the posterior mean from the cached value of FITC kernels

#         Parameters
#         ----------
#         input: torch.Tensor(N_xin, D)
#             input torch Tensor

#         Returns
#         -------
#         res: torch.Tensor(N_xin)
#             posterior mean value of FITC GP model
#         """
#         # detach is applied here to avoid updating the same parameter twice in the same iteration
#         # which might result in an error
#         res = self.covar_module.posterior_mean(input).detach()
#         return res



from __future__ import annotations

import numpy as np

from smac import constants
from smac.runhistory.encoder.encoder import RunHistoryEncoder
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


logger = get_logger(__name__)


class RunHistoryScaledEncoder(RunHistoryEncoder):
    def transform_response_values(self, values: np.ndarray) -> np.ndarray:
        """Transforms the response values by linearly scaling them between zero and one."""
        min_y = self._min_y - (
            self._percentile - self._min_y
        )  # Subtract the difference between the percentile and the minimum
        min_y -= constants.VERY_SMALL_NUMBER  # Minimal value to avoid numerical issues in the log scaling below

        # Linear scaling
        # prevent diving by zero
        min_y[np.where(min_y == self._max_y)] *= 1 - 10**-101
        values = (values - min_y) / (self._max_y - min_y)
        return values



from __future__ import annotations

# We don't want to create a "real" package here so we just work with this hack
import sys

sys.path.insert(0, ".")

import logging

logging.disable(9999)

from typing import Any

import numpy as np
import json
import socket  # noqa: E402
from collections import defaultdict
from pathlib import Path

from smac.utils.numpyencoder import NumpyEncoder

import pandas as pd
from src.tasks import TASKS  # noqa: E402
from src.utils.exceptions import NotSupportedError  # noqa: E402
from src.utils.styled_plot import plt  # noqa: E402
from src.wrappers.v14 import Version14  # noqa: E402
from src.wrappers.v20 import Version20  # noqa: E402
from src.wrappers.wrapper import Wrapper  # noqa: E402

SEEDS = [0, 50, 100, 150, 200] # , 250, 300, 350, 400, 450]
WRAPPERS = [Version14, Version20]
RAW_FILENAME = Path("report/raw.json")


class Benchmark:
    """Selects the right wrapper (based on the environment), runs, and saves the src."""

    def __init__(self, overwrite: bool = False) -> None:
        import smac

        self._version: str | None = None
        for version in ["__version__", "version"]:
            if hasattr(smac, version):
                version = getattr(smac, version)
                break

        if version is None:
            raise RuntimeError("Could not find version of SMAC.")
        else:
            version = version.replace("v", "")

        self._wrapper: Wrapper | None = None
        for wrapper in WRAPPERS:
            if version in wrapper.supported_versions:
                self._wrapper = wrapper
                break

        if self._wrapper is None:
            raise RuntimeError(f"Could not find a wrapper for version {version}.")

        print(f"Found wrapper for version {version}.")

        self._version: str = version
        self._computer: str = str(socket.gethostname())
        self._data: Any = {}
        self._overwrite: bool = overwrite

    def _load_data(self) -> None:
        """Loads data from the file to object cache."""
        print("Loading data...")
        RAW_FILENAME.parent.mkdir(parents=True, exist_ok=True)

        # First, we load all the previous data
        if RAW_FILENAME.exists() and not self._overwrite:
            with open(RAW_FILENAME, "r") as f:
                self._data = json.load(f)
        else:
            self._data = {}

    def _save_data(self) -> None:
        """Saves the internal data to the file."""
        print("Saving data...")
        with open(str(RAW_FILENAME), "w") as f:
            json.dump(self._data, f, indent=4, cls=NumpyEncoder)

    def _fill_keys(self) -> None:
        """Fill data with keys based on computer name, tasks, and selected version."""
        print("Filling keys...")
        if self._computer not in self._data:
            self._data[self._computer] = {}

        for task in TASKS:
            if task.id not in self._data[self._computer]:
                self._data[self._computer][task.id] = {}

            if "name" not in self._data[self._computer][task.id]:
                self._data[self._computer][task.id]["name"] = task.name

            if "objective" not in self._data[self._computer][task.id]:
                self._data[self._computer][task.id]["objective"] = task.objectives[0]

            if "trajectory_trials" not in self._data[self._computer][task.id]:
                self._data[self._computer][task.id]["trajectory_trials"] = {}

            if "trajectory_walltime" not in self._data[self._computer][task.id]:
                self._data[self._computer][task.id]["trajectory_walltime"] = {}

            if self._version not in self._data[self._computer][task.id]["trajectory_trials"]:
                self._data[self._computer][task.id]["trajectory_trials"][self._version] = {}

            if self._version not in self._data[self._computer][task.id]["trajectory_walltime"]:
                self._data[self._computer][task.id]["trajectory_walltime"][self._version] = {}

        self._save_data()

    def _get_mean_std(self, values: list[tuple[list[float], list[float]]]) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        # First we have to get all x values
        X: list[float] = []
        for x_values, _ in values:
            for x in x_values:
                if x not in X:
                    X.append(x)

        # Let's sort the x values
        X.sort()

        if len(X) == 0:
            return np.array([]), np.array([]), np.array([])

        # Now we need to look for corresponding y values
        # Each x is associated with #seed y values
        Y: list[list[float]] = []
        for x in X:
            y = []

            # Iterate over the values again
            for x_values, y_values in values:
                # Findest closest value to x
                idx = min(range(len(x_values)), key=lambda i: abs(x_values[i] - x))
                y.append(y_values[idx])

            Y.append(y)

        # Make numpy arrays
        Y_array = np.array(Y)

        mean = np.mean(Y_array, axis=1)
        std = np.std(Y_array, axis=1)

        return np.array(X), Y_array, mean, std

    def run(self) -> None:
        # Get name of the current computer (for comparison purposes)
        assert self._wrapper is not None

        # First, we load all the previous data
        self._load_data()

        # Fill keys of the data object
        self._fill_keys()

        # Now iterate over all tasks
        for task in TASKS:
            for seed in SEEDS:
                print(f"--- Start {task.name} with seed {seed}")

                # Some variables
                d = self._data[self._computer][task.id]

                # Check if we already calculated the data before
                if (
                    str(seed) in d["trajectory_trials"][self._version]
                    and str(seed) in d["trajectory_walltime"][self._version]
                ):
                    print(f"Already done {task.name}. Skipping.")
                    continue
                else:
                    print("Not found in the data. Running...")

                task_wrapper = self._wrapper(task, seed)
                try:
                    task_wrapper.run()
                except NotSupportedError:
                    print("Not supported. Skipping.")
                    continue
                except Exception as e:
                    print("Something went wrong:")
                    print(e)
                    continue

                for sort_by in ["trials", "walltime"]:
                    # Now we collect some metrics
                    try:
                        X, Y = task_wrapper.get_trajectory(sort_by=sort_by)
                    except NotSupportedError:
                        continue

                    d[f"trajectory_{sort_by}"][self._version][str(seed)] = (X, Y)

                self._save_data()

        # Plot things
        self._plot_trajectory()

        # Write things
        self._write_table()

    def _plot_trajectory(self) -> None:
        import seaborn as sns
        sns.set_style("whitegrid")
        sns.set_palette("colorblind")

        d = self._data[self._computer]

        for sort_by in ["trials", "walltime"]:
            len_tasks = len(d.keys())

            traj = f"trajectory_{sort_by}"
            filename = Path(f"report/{traj}.png")

            

            # Collect data
            df = []
            for task in TASKS:
                objective = d[task.id]["objective"]
                
                for version, seeds in d[task.id][traj].items():
                    for seed, (X, Y) in seeds.items():
                        X = np.array(X)
                        Y = np.array(Y)
                        df.append(pd.DataFrame({
                            "task": task.name,
                            "objective": objective,
                            "version": [version] * len(X),
                            "seed": [seed] * len(Y),
                            "x": X,
                            "y": Y,
                        }))
                        # assert np.all(X[1:] > X[:-1]), f"X not strictly monotonically increasing, {version, seed, X, Y}"
                        # assert np.all(Y[1:] < Y[:-1]), f"Y not strictly monotonically decreasing, {version, seed, X, Y}"
            plot_df = pd.concat(df).reset_index(drop=True)

            # Plot
            plt.figure(rows=len_tasks)
            plt.tight_layout()
            plt.subplots_adjust(wspace=0.6)

            hue_order = list(plot_df["version"].unique())

            i = 0
            for task in TASKS:
                objective = d[task.id]["objective"]

                plt.subplot(len_tasks, 1, i + 1)
                plt.title(task.name)

                df = plot_df[plot_df["task"] == task.name]

                # Fill missing values
                x_unique = df["x"].unique()
                new_df = []
                for gid, gdf in df.groupby(by=["version", "seed"]):
                    x_missing = list(set(x_unique).difference(set(gdf["x"].unique())))
                    gdf = pd.concat((gdf, 
                        pd.DataFrame({
                            "version": gid[0],
                            "seed": gid[1],
                            "x": x_missing,
                            "y": [np.nan] * len(x_missing),
                        })
                    ))
                    gdf.sort_values(by="x", inplace=True)
                    gdf.ffill(inplace=True)  # forward fill
                    new_df.append(gdf)
                df = pd.concat(new_df).reset_index(drop=True)

                ax = sns.lineplot(data=df, x="x", y="y", hue="version", style=None, lw=1, hue_order=hue_order)
                ax.legend()
                # ax.get_legend().remove()

                # plt.legend()
                ax.set_ylabel(objective)

                if task.y_log_scale:
                    ax.set_yscale("log")

                if task.x_log_scale:
                    ax.set_xscale("log")

                if i == len_tasks - 1:
                    ax.set_xlabel(sort_by)

                i += 1

            plt.save_figure(str(filename))

    def _write_table(self) -> None:
        """Writes a table with the best result found so far."""
        d = self._data[self._computer]

        versions = set()
        for task in TASKS:
            for version in d[task.id]["trajectory_walltime"].keys():
                versions.add(version)

        # For each version a column
        data = defaultdict(list)
        for task in TASKS:
            data["tasks"].append(task.name)
            data["n_trials"].append(str(task.n_trials))
            data["walltime"].append(str(task.walltime_limit))
            data["instances"].append(str(task.use_instances))

            for version in versions:
                not_found = False
                for seed in SEEDS:
                    if seed not in d[task.id]["trajectory_walltime"][version]:
                        not_found = True
                        break

                if not_found:
                    data[f"{version} (cost / time)"].append("N/A")
                    continue

                values = d[task.id]["trajectory_walltime"][version].values()
                X, Y_array, Y_mean, Y_std = self._get_mean_std(values)
                data[f"{version} (cost / time)"].append(
                    f"{round(Y_mean[-1], 2)} +- {round(Y_std[-1], 2)} / {round(X[-1], 2)}"
                )

        # Now make a dataframe
        df = pd.DataFrame(data=data)
        df.to_csv("report/table.csv", index=False)


if __name__ == "__main__":
    benchmark = Benchmark(overwrite=False)
    benchmark.run()



from __future__ import annotations

from typing import Any, Optional, TypeVar, cast

import logging

import numpy as np
from ConfigSpace import ConfigurationSpace
from scipy import optimize
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Kernel

from smac.constants import VERY_SMALL_NUMBER
from smac.model.gaussian_process.abstract_gaussian_process import (
    AbstractGaussianProcess,
)
from smac.model.gaussian_process.priors.abstract_prior import AbstractPrior

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


logger = logging.getLogger(__name__)


Self = TypeVar("Self", bound="GaussianProcess")


class GaussianProcess(AbstractGaussianProcess):
    """Implementation of Gaussian process model. The Gaussian process hyperparameters are obtained by optimizing
    the marginal log likelihood.

    This code is based on the implementation of RoBO:
    Klein, A. and Falkner, S. and Mansur, N. and Hutter, F.
    RoBO: A Flexible and Robust Bayesian Optimization Framework in Python
    In: NIPS 2017 Bayesian Optimization Workshop

    Parameters
    ----------
    configspace : ConfigurationSpace
    kernel : Kernel
        Kernel which is used for the Gaussian process.
    n_restarts : int, defaults to 10
        Number of restarts for the Gaussian process hyperparameter optimization.
    normalize_y : bool, defaults to True
        Zero mean unit variance normalization of the output values.
    instance_features : dict[str, list[int | float]] | None, defaults to None
        Features (list of int or floats) of the instances (str). The features are incorporated into the X data,
        on which the model is trained on.
    pca_components : float, defaults to 7
        Number of components to keep when using PCA to reduce dimensionality of instance features.
    seed : int
    """

    def __init__(
        self,
        configspace: ConfigurationSpace,
        kernel: Kernel,
        n_restarts: int = 10,
        normalize_y: bool = True,
        instance_features: dict[str, list[int | float]] | None = None,
        pca_components: int | None = 7,
        seed: int = 0,
    ):
        super().__init__(
            configspace=configspace,
            seed=seed,
            kernel=kernel,
            instance_features=instance_features,
            pca_components=pca_components,
        )

        self._normalize_y = normalize_y
        self._n_restarts = n_restarts

        # Internal variables
        self._hypers = np.empty((0,))
        self._is_trained = False
        self._n_ll_evals = 0

        self._set_has_conditions()

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update({"n_restarts": self._n_restarts, "normalize_y": self._normalize_y})

        return meta

    def _train(
        self: Self,
        X: np.ndarray,
        y: np.ndarray,
        optimize_hyperparameters: bool = True,
    ) -> Self:
        """Computes the Cholesky decomposition of the covariance of X and estimates the GP
        hyperparameters by optimizing the marginal log likelihood. The prior mean of the GP is set to
        the empirical mean of X.

        Parameters
        ----------
        X : np.ndarray [#samples, #hyperparameters + #features]
            Input data points.
        Y : np.ndarray [#samples, #objectives]
            The corresponding target values.
        optimize_hyperparameters: boolean
            If set to true, the hyperparameters are optimized, otherwise the default hyperparameters of the kernel are
            used.
        """
        if self._normalize_y:
            y = self._normalize(y)

        X = self._impute_inactive(X)
        y = y.flatten()

        n_tries = 10
        for i in range(n_tries):
            try:
                self._gp = self._get_gaussian_process()
                self._gp.fit(X, y)
                break
            except np.linalg.LinAlgError as e:
                if i == n_tries:
                    raise e

                # Assume that the last entry of theta is the noise
                theta = np.exp(self._kernel.theta)
                theta[-1] += 1
                self._kernel.theta = np.log(theta)

        if optimize_hyperparameters:
            self._all_priors = self._get_all_priors(add_bound_priors=False)
            self._hypers = self._optimize()
            self._gp.kernel.theta = self._hypers
            self._gp.fit(X, y)
        else:
            self._hypers = self._gp.kernel.theta

        # Set the flag
        self._is_trained = True

        return self

    def _get_gaussian_process(self) -> GaussianProcessRegressor:
        return GaussianProcessRegressor(
            kernel=self._kernel,
            normalize_y=False,  # We do not use scikit-learn's normalize routine
            optimizer=None,
            n_restarts_optimizer=0,  # We do not use scikit-learn's optimization routine
            alpha=0,  # Governed by the kernel
            random_state=self._rng,
        )

    def _nll(self, theta: np.ndarray) -> tuple[float, np.ndarray]:
        """Returns the negative marginal log likelihood (+ the prior) for a hyperparameter
        configuration theta. Negative because we use scipy minimize for optimization.

        Parameters
        ----------
        theta : np.ndarray
            Hyperparameter vector. Note that all hyperparameter are on a log scale.
        """
        self._n_ll_evals += 1

        try:
            lml, grad = self._gp.log_marginal_likelihood(theta, eval_gradient=True)
        except np.linalg.LinAlgError:
            return 1e25, np.zeros(theta.shape)

        for dim, priors in enumerate(self._all_priors):
            for prior in priors:
                lml += prior.get_log_probability(theta[dim])
                grad[dim] += prior.get_gradient(theta[dim])

        # We add a minus here because scipy is minimizing
        if not np.isfinite(lml).all() or not np.all(np.isfinite(grad)):
            return 1e25, np.zeros(theta.shape)
        else:
            return -lml, -grad

    def _optimize(self) -> np.ndarray:
        """Optimizes the marginal log likelihood and returns the best found hyperparameter
        configuration theta.

        Returns
        -------
        theta : np.ndarray
            Hyperparameter vector that maximizes the marginal log likelihood.
        """
        log_bounds = [(b[0], b[1]) for b in self._gp.kernel.bounds]

        # Start optimization from the previous hyperparameter configuration
        p0 = [self._gp.kernel.theta]
        if self._n_restarts > 0:
            dim_samples = []

            prior: list[AbstractPrior] | AbstractPrior | None = None
            for dim, hp_bound in enumerate(log_bounds):
                prior = self._all_priors[dim]
                # Always sample from the first prior
                if isinstance(prior, list):
                    if len(prior) == 0:
                        prior = None
                    else:
                        prior = prior[0]

                prior = cast(Optional[AbstractPrior], prior)
                if prior is None:
                    try:
                        sample = self._rng.uniform(
                            low=hp_bound[0],
                            high=hp_bound[1],
                            size=(self._n_restarts,),
                        )
                    except OverflowError:
                        raise ValueError("OverflowError while sampling from (%f, %f)" % (hp_bound[0], hp_bound[1]))

                    dim_samples.append(sample.flatten())
                else:
                    dim_samples.append(prior.sample_from_prior(self._n_restarts).flatten())
            p0 += list(np.vstack(dim_samples).transpose())

        theta_star: np.ndarray | None = None
        f_opt_star = np.inf
        for i, start_point in enumerate(p0):
            theta, f_opt, _ = optimize.fmin_l_bfgs_b(self._nll, start_point, bounds=log_bounds)
            if f_opt < f_opt_star:
                f_opt_star = f_opt
                theta_star = theta

        if theta_star is None:
            raise RuntimeError

        return theta_star

    def _predict(
        self,
        X: np.ndarray,
        covariance_type: str | None = "diagonal",
    ) -> tuple[np.ndarray, np.ndarray | None]:
        if not self._is_trained:
            raise Exception("Model has to be trained first!")

        X_test = self._impute_inactive(X)

        if covariance_type is None:
            mu = self._gp.predict(X_test)
            var = None

            if self._normalize_y:
                mu = self._untransform_y(mu)
        else:
            predict_kwargs = {"return_cov": False, "return_std": True}
            if covariance_type == "full":
                predict_kwargs = {"return_cov": True, "return_std": False}

            mu, var = self._gp.predict(X_test, **predict_kwargs)

            if covariance_type != "full":
                var = var**2  # Since we get standard deviation for faster computation

            # Clip negative variances and set them to the smallest
            # positive float value
            var = np.clip(var, VERY_SMALL_NUMBER, np.inf)

            if self._normalize_y:
                mu, var = self._untransform_y(mu, var)

            if covariance_type == "std":
                var = np.sqrt(var)  # Converting variance to std deviation if specified

        return mu, var

    def sample_functions(self, X_test: np.ndarray, n_funcs: int = 1) -> np.ndarray:
        """Samples F function values from the current posterior at the N specified test points.

        Parameters
        ----------
        X : np.ndarray [#samples, #hyperparameters + #features]
            Input data points.
        n_funcs: int
            Number of function values that are drawn at each test point.

        Returns
        -------
        function_samples : np.ndarray
            The F function values drawn at the N test points.
        """
        if not self._is_trained:
            raise Exception("Model has to be trained first.")

        X_test = self._impute_inactive(X_test)
        funcs = self._gp.sample_y(X_test, n_samples=n_funcs, random_state=self._rng)

        if self._normalize_y:
            funcs = self._untransform_y(funcs)

        if len(funcs.shape) == 1:
            return funcs[None, :]
        else:
            return funcs



from __future__ import annotations

import numpy as np
import scipy.optimize
import scipy.spatial.distance
import scipy.special
import sklearn.gaussian_process.kernels as kernels

from smac.model.gaussian_process.kernels.base_kernels import AbstractKernel
from smac.model.gaussian_process.priors.abstract_prior import AbstractPrior

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class RBFKernel(AbstractKernel, kernels.RBF):
    """RBF kernel implementation."""

    def __init__(
        self,
        length_scale: float | tuple[float, ...] | np.ndarray = 1.0,
        length_scale_bounds: tuple[float, float] | list[tuple[float, float]] | np.ndarray = (1e-5, 1e5),
        operate_on: np.ndarray | None = None,
        has_conditions: bool = False,
        prior: AbstractPrior | None = None,
    ) -> None:
        super().__init__(
            operate_on=operate_on,
            has_conditions=has_conditions,
            prior=prior,
            length_scale=length_scale,
            length_scale_bounds=length_scale_bounds,
        )

    def _call(
        self,
        X: np.ndarray,
        Y: np.ndarray | None = None,
        eval_gradient: bool = False,
        active: np.ndarray | None = None,
    ) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
        X = np.atleast_2d(X)
        length_scale = kernels._check_length_scale(X, self.length_scale)

        if Y is None:
            dists = scipy.spatial.distance.pdist(X / length_scale, metric="sqeuclidean")
            K = np.exp(-0.5 * dists)
            # convert from upper-triangular matrix to square matrix
            K = scipy.spatial.distance.squareform(K)
            np.fill_diagonal(K, 1)
        else:
            if eval_gradient:
                raise ValueError("Gradient can only be evaluated when Y is None.")
            dists = scipy.spatial.distance.cdist(X / length_scale, Y / length_scale, metric="sqeuclidean")
            K = np.exp(-0.5 * dists)

        if active is not None:
            K = K * active

        if eval_gradient:
            if self.hyperparameter_length_scale.fixed:
                # Hyperparameter l kept fixed
                return K, np.empty((X.shape[0], X.shape[0], 0))
            elif not self.anisotropic or length_scale.shape[0] == 1:
                K_gradient = (K * scipy.spatial.distance.squareform(dists))[:, :, np.newaxis]
                return K, K_gradient
            elif self.anisotropic:
                # We need to recompute the pairwise dimension-wise distances
                K_gradient = (X[:, np.newaxis, :] - X[np.newaxis, :, :]) ** 2 / (length_scale**2)
                K_gradient *= K[..., np.newaxis]
                return K, K_gradient

        return K



from __future__ import annotations

from typing import Any

import numpy as np

from smac.multi_objective.abstract_multi_objective_algorithm import (
    AbstractMultiObjectiveAlgorithm,
)
from smac.scenario import Scenario


class MeanAggregationStrategy(AbstractMultiObjectiveAlgorithm):
    """A class to mean-aggregate multi-objective costs to a single cost.

    Parameters
    ----------
    scenario : Scenario
    objective_weights : list[float] | None, defaults to None
        Weights for an weighted average. Must be of the same length as the number of objectives.
    """

    def __init__(
        self,
        scenario: Scenario,
        objective_weights: list[float] | None = None,
    ):
        super(MeanAggregationStrategy, self).__init__()

        if objective_weights is not None and scenario.count_objectives() != len(objective_weights):
            raise ValueError("Number of objectives and number of weights must be equal.")

        self._objective_weights = objective_weights

    @property
    def meta(self) -> dict[str, Any]:
        """Returns the meta data of the created object."""
        return {
            "name": self.__class__.__name__,
            "objective_weights": self._objective_weights,
        }

    def __call__(self, values: list[float]) -> float:  # noqa: D102
        return float(np.average(values, axis=0, weights=self._objective_weights))



import os

import setuptools

from smac import (
    author,
    author_email,
    description,
    package_name,
    project_urls,
    url,
    version,
)

HERE = os.path.dirname(os.path.realpath(__file__))


def read_file(filepath: str) -> str:
    with open(filepath, "r", encoding="utf-8") as fh:
        return fh.read()


extras_require = {
    "dev": [
        "setuptools",
        "types-setuptools",
        # Test
        "pytest>=4.6",
        "pytest-cov",
        "pytest-xdist",
        "pytest-timeout",
        # Docs
        "automl-sphinx-theme>=0.2",
        # Others
        "mypy",
        "isort",
        "black",
        "pydocstyle",
        "flake8",
        "pre-commit",
        "pylint",
        "mkdocs",
        "mkdocs-material",
        "mkdocs-autorefs",
        "mkdocs-gen-files",
        "mkdocs-literate-nav",
        "mkdocs-glightbox",
        "mkdocs-glossary-plugin",
        "mkdocstrings[python]",
        "markdown-exec[ansi]",
        "mike",
        "pillow",
        "cairosvg",
        "black",                # This allows mkdocstrings to format signatures in the docs
        "pytest",
        "pytest-coverage",
        "pytest-cases",
    ],
}

setuptools.setup(
    name=package_name,
    author=author,
    author_email=author_email,
    description=description,
    long_description=read_file(os.path.join(HERE, "README.md")),
    long_description_content_type="text/markdown",
    license="BSD 3-Clause License",
    url=url,
    project_urls=project_urls,
    version=version,
    packages=setuptools.find_packages(exclude=["tests"]),
    include_package_data=True,
    python_requires=">=3.8",
    install_requires=[
        "numpy",
        "scipy>=1.9.2",
        "psutil",
        "pynisher>=1.0.0",
        "ConfigSpace>=1.0.0",
        "joblib",
        "scikit-learn>=1.1.2",
        "pyrfr>=0.9.0",
        "dask[distributed]",
        "dask_jobqueue>=0.8.2",
        "emcee>=3.0.0",
        "regex",
        "pyyaml",
    ],
    extras_require=extras_require,
    test_suite="pytest",
    platforms=["Linux"],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Natural Language :: English",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



from __future__ import annotations

from typing import Any

import time
from subprocess import PIPE, Popen

from ConfigSpace import Configuration

from smac.runner.abstract_runner import StatusType
from smac.runner.abstract_serial_runner import AbstractSerialRunner
from smac.scenario import Scenario
from smac.utils.logging import get_logger

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class TargetFunctionScriptRunner(AbstractSerialRunner):
    """Class to execute target functions from scripts. Uses `Popen` to execute the script in a
     subprocess.

    The following example shows how the script is called:
    ``target_function --instance=test --instance_features=test --seed=0 --hyperparameter1=5323``

    The script must return an echo in the following form (white-spaces are removed):
    ``cost=0.5; runtime=0.01; status=SUCCESS; additional_info=test`` (single-objective)
    ``cost=0.5, 0.4; runtime=0.01; status=SUCCESS; additional_info=test`` (multi-objective)

    The status must be a string and must be one of the ``StatusType`` values. However, ``runtime``,
    ``status`` and ``additional_info`` are optional.

    Note
    ----
    Everytime an instance is passed, also an instance feature in form of a comma-separated list
    (no spaces) of floats is passed. If no instance feature for the instance is given,
    an empty list is passed.

    Parameters
    ----------
    target_function : Callable
        The target function.
    scenario : Scenario
    required_arguments : list[str]
        A list of required arguments, which are passed to the target function.
    """

    def __init__(
        self,
        target_function: str,
        scenario: Scenario,
        required_arguments: list[str] = None,
    ):
        if required_arguments is None:
            required_arguments = []
        super().__init__(scenario=scenario, required_arguments=required_arguments)
        self._target_function = target_function

        # Check if target function is callable
        if not isinstance(self._target_function, str):
            raise TypeError(
                "Argument `target_function` must be a string but is type" f"`{type(self._target_function)}`."
            )

        if self._scenario.trial_memory_limit is not None:
            logger.warning("Trial memory limit is not supported for script target functions.")

        if self._scenario.trial_walltime_limit is not None:
            logger.warning("Trial walltime limit is not supported for script target functions.")

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update({"filename": str(self._target_function)})

        return meta

    def run(
        self,
        config: Configuration,
        instance: str | None = None,
        budget: float | None = None,
        seed: int | None = None,
    ) -> tuple[StatusType, float | list[float], float, float, dict]:
        """Calls the target function.

        Parameters
        ----------
        config : Configuration
            Configuration to be passed to the target function.
        instance : str | None, defaults to None
            The Problem instance.
        budget : float | None, defaults to None
            A positive, real-valued number representing an arbitrary limit to the target function
            handled by the target function internally.
        seed : int, defaults to None

        Returns
        -------
        status : StatusType
            Status of the trial.
        cost : float | list[float]
            Resulting cost(s) of the trial.
        runtime : float
            The time the target function took to run.
        cpu_time : float
            The time the target function took on the hardware to run.
        additional_info : dict
            All further additional trial information.
        """
        # The kwargs are passed to the target function.
        kwargs: dict[str, Any] = {}
        if "seed" in self._required_arguments:
            kwargs["seed"] = seed

        if "instance" in self._required_arguments:
            kwargs["instance"] = instance

            # In contrast to the normal target function runner, we also add the instance features here.
            if self._scenario.instance_features is not None and instance in self._scenario.instance_features:
                kwargs["instance_features"] = self._scenario.instance_features[instance]
            else:
                kwargs["instance_features"] = []

        if "budget" in self._required_arguments:
            kwargs["budget"] = budget

        # Presetting
        cost: float | list[float] = self._crash_cost
        runtime = 0.0
        cpu_time = runtime
        additional_info = {}
        status = StatusType.SUCCESS

        # Add config arguments to the kwargs
        for k, v in dict(config).items():
            if k in kwargs:
                raise RuntimeError(f"The key {k} is already in use. Please use a different one.")
            kwargs[k] = v

        # Call target function
        start_time = time.time()
        cpu_time = time.process_time()
        output, error = self(kwargs)
        cpu_time = time.process_time() - cpu_time
        runtime = time.time() - start_time

        # Now we have to parse the std output
        # First remove white-spaces
        output = output.replace(" ", "")

        outputs = {}
        for pair in output.split(";"):
            try:
                kv = pair.split("=")
                k, v = kv[0], kv[1]

                # Get rid of the trailing newline
                v = v.strip()

                outputs[k] = v
            except Exception:
                pass

        # Parse status
        if "status" in outputs:
            status = getattr(StatusType, outputs["status"])

        # Parse costs (depends on the number of objectives)
        if "cost" in outputs:
            if self._n_objectives == 1:
                cost = float(outputs["cost"])
            else:
                costs = outputs["cost"].split(",")
                costs = [float(c) for c in costs]
                cost = costs

                if len(costs) != self._n_objectives:
                    raise RuntimeError("The number of costs does not match the number of objectives.")
        else:
            status = StatusType.CRASHED

        # Overwrite runtime
        if "runtime" in outputs:
            runtime = float(outputs["runtime"])

        # Overwrite CPU time
        if "cpu_time" in outputs:
            cpu_time = float(outputs["cpu_time"])

        # Add additional info
        if "additional_info" in outputs:
            additional_info["additional_info"] = outputs["additional_info"]

        if status != StatusType.SUCCESS:
            additional_info["error"] = error

            if cost != self._crash_cost:
                cost = self._crash_cost
                logger.info(
                    "The target function crashed but returned a cost. The cost is ignored and replaced by crash cost."
                )

        return status, cost, runtime, cpu_time, additional_info

    def __call__(
        self,
        algorithm_kwargs: dict[str, Any],
    ) -> tuple[str, str]:
        """Calls the algorithm, which is processed in the ``run`` method."""
        cmd = [self._target_function]
        for k, v in algorithm_kwargs.items():
            v = str(v)
            k = str(k)

            # Let's remove some spaces
            v = v.replace(" ", "")

            cmd += [f"--{k}={v}"]

        logger.debug(f"Calling: {' '.join(cmd)}")
        p = Popen(cmd, shell=False, stdout=PIPE, stderr=PIPE, universal_newlines=True)
        output, error = p.communicate()

        logger.debug("Stdout: %s" % output)
        logger.debug("Stderr: %s" % error)

        return output, error



from __future__ import annotations


def normalize_costs(
    values: list[float],
    bounds: list[tuple[float, float]] | None = None,
) -> list[float]:
    """
    Normalizes a list of floats with corresponding bounds.

    Parameters
    ----------
    values : list[float]
        List of costs to be normalized.
    bounds : list[tuple[float, float]] | None, optional, defaults to None
        List of tuple of bounds. If no bounds are passed, the input is returned.

    Returns
    -------
    normalized_costs : list[float]
        Normalized costs based on the bounds. If no bounds are given, the original values are returned.
        Also, if min and max bounds are the same, the value of the corresponding objective is set to 1.
    """
    if bounds is None:
        return values

    if len(values) != len(bounds):
        raise ValueError("Number of values and bounds must be equal.")

    costs = []
    for v, b in zip(values, bounds):
        assert not isinstance(v, list)
        p = v - b[0]
        q = b[1] - b[0]

        if q < 1e-10:
            cost = 1.0
        else:
            cost = p / q
        costs.append(cost)

    return costs



# from __future__ import annotations

# from abc import ABC, abstractmethod
# from typing import Any, Dict, Iterator, List, Tuple, Type

# import copy
# import inspect
# import logging
# import math

# import numpy as np
# from ConfigSpace.forbidden import (
#     AbstractForbiddenComponent,
#     ForbiddenAndConjunction,
#     MultipleValueForbiddenClause,
# )
# from ConfigSpace.hyperparameters import (
#     CategoricalHyperparameter,
#     Constant,
#     Hyperparameter,
#     NumericalHyperparameter,
#     OrdinalHyperparameter,
#     UniformFloatHyperparameter,
#     UniformIntegerHyperparameter,
# )

# from smac.acquisition.function import EI, AbstractAcquisitionFunction
# from ConfigSpace import Configuration, ConfigurationSpace
# from smac.model.abstract_model import AbstractModel
# from smac.model.gaussian_process.gpytorch_gaussian_process import GloballyAugmentedLocalGaussianProcess
# from smac.model.gaussian_process.kernels._boing import construct_gp_kernel
# from smac.model.utils import check_subspace_points
# from smac.utils.logging import get_logger

# logger = get_logger(__name__)


# class LocalSubspace(ABC):
#     """
#     A subspace that is designed for local Bayesian Optimization. If bounds_ss_cont and bounds_ss_cat are not given,
#     this subspace is equivalent to the original configuration space. Additionally, this subspace
#     supports local BO that only works with a subset of the dimensions, where the missing values are filled by the
#     corresponding values from incumbent_array.

#     Parameters
#     ----------
#     config_space: ConfigurationSpace
#         raw Configuration space
#     bounds: List[Tuple[float, float]]
#         raw bounds of the Configuration space, notice that here bounds denotes the bounds of the entire space
#     hps_types: List[int],
#         types of the hyperparameters
#     bounds_ss_cont: np.ndarray(D_cont, 2)
#         subspaces bounds of continuous hyperparameters, its length is the number of continuous hyperparameters
#     bounds_ss_cat: List[Tuple]
#         subspaces bounds of categorical hyperparameters, its length is the number of categorical hyperparameters
#     rng: np.random.RandomState
#         random state
#     model_local: ~smac.epm.base_epm.BaseEPM
#         model in subspace
#     model_local_kwargs: Dict | None
#         argument for subspace model
#     acq_func_local: ~smac.optimizer.ei_optimization.AbstractAcquisitionFunction
#         local acquisition function
#     acq_func_local_kwargs: Dict | None
#         argument for acquisition function
#     activate_dims: np.ndarray | None
#         activate dimensions in the subspace, if it is None, we preserve all the dimensions
#     incumbent_array: np.ndarray | None
#         incumbent array, used when activate_dims has less dimension and this value is used to complementary the
#         resulted configurations
#     """

#     def __init__(
#         self,
#         config_space: ConfigurationSpace,
#         bounds: List[Tuple[float, float]],
#         hps_types: List[int],
#         bounds_ss_cont: np.ndarray | None = None,
#         bounds_ss_cat: List[Tuple] | None = None,
#         model_local: AbstractModel | Type[AbstractModel] = GloballyAugmentedLocalGaussianProcess,
#         model_local_kwargs: Dict = {},
#         acq_func_local: AbstractAcquisitionFunction | Type[AbstractAcquisitionFunction] = EI,
#         acq_func_local_kwargs: Dict | None = None,
#         rng: np.random.RandomState | None = None,
#         initial_data: Tuple[np.ndarray, np.ndarray] | None = None,
#         activate_dims: np.ndarray | None = None,
#         incumbent_array: np.ndarray | None = None,
#     ):
#         self.cs_global = config_space
#         if rng is None:
#             self.rng = np.random.RandomState(1)
#         else:
#             self.rng = np.random.RandomState(rng.randint(0, 2**20))

#         n_hypers = len(config_space.get_hyperparameters())
#         model_types = copy.deepcopy(hps_types)
#         model_bounds = copy.deepcopy(bounds)

#         cat_dims = np.where(np.array(hps_types) != 0)[0]
#         cont_dims = np.where(np.array(hps_types) == 0)[0]

#         if activate_dims is None:
#             activate_dims = np.arange(n_hypers)
#             activate_dims_cont = cont_dims
#             activate_dims_cat = cat_dims
#             self.activate_dims = activate_dims
#             activate_dims_cont_ss = np.arange(len(activate_dims_cont))
#             activate_dims_cat_ss = np.arange(len(activate_dims_cat))
#         else:
#             activate_dims_cont, _, activate_dims_cont_ss = np.intersect1d(
#                 activate_dims, cont_dims, assume_unique=True, return_indices=True
#             )
#             activate_dims_cat, _, activate_dims_cat_ss = np.intersect1d(
#                 activate_dims, cat_dims, assume_unique=True, return_indices=True
#             )
#             self.activate_dims = activate_dims

#         self.activate_dims_cont = activate_dims_cont_ss
#         self.activate_dims_cat = activate_dims_cat_ss

#         lbs = np.full(n_hypers, 0.0)
#         scales = np.full(n_hypers, 1.0)

#         if bounds_ss_cont is None and bounds_ss_cat is None:
#             # cs_inner is cs
#             self.cs_local = config_space
#             self.new_config_space = False
#             self.bounds_ss_cont = np.tile([0.0, 1.0], [len(self.activate_dims_cont), 1])
#             self.bounds_ss_cat = []  # type: Optional[List[Tuple]]
#             self.lbs = lbs
#             self.scales = scales
#             self.new_config = False

#         else:
#             self.new_config = True
#             # we normalize the non-CategoricalHyperparameter by x = (x-lb)*scale

#             hps = config_space.get_hyperparameters()

#             # deal with categorical hyperaprameters
#             for i, cat_idx in enumerate(activate_dims_cat):
#                 hp_cat = hps[cat_idx]  # type: CategoricalHyperparameter
#                 parents = config_space.get_parents_of(hp_cat.name)
#                 if len(parents) == 0:
#                     can_be_inactive = False
#                 else:
#                     can_be_inactive = True
#                 if bounds_ss_cat is None:
#                     n_cats = len(hp_cat.choices)
#                 else:
#                     n_cats = len(bounds_ss_cat[i])
#                 if can_be_inactive:
#                     n_cats = n_cats + 1
#                 model_types[cat_idx] = n_cats
#                 model_bounds[cat_idx] = (int(n_cats), np.nan)

#             # store the dimensions of numerical hyperparameters, UniformFloatHyperparameter and
#             # UniformIntegerHyperparameter
#             dims_cont_num = []
#             idx_cont_num = []
#             dims_cont_ord = []
#             idx_cont_ord = []
#             ord_hps = {}

#             # deal with ordinary hyperaprameters
#             for i, cont_idx in enumerate(activate_dims_cont):
#                 param = hps[cont_idx]
#                 if isinstance(param, OrdinalHyperparameter):
#                     parents = config_space.get_parents_of(param.name)
#                     if len(parents) == 0:
#                         can_be_inactive = False
#                     else:
#                         can_be_inactive = True
#                     if bounds_ss_cont is None:
#                         n_seqs = len(param.sequence)
#                     else:
#                         n_seqs = bounds_ss_cont[i][1] - bounds_ss_cont[i][0] + 1
#                     if can_be_inactive:
#                         model_bounds[cont_idx] = (0, int(n_seqs))
#                     else:
#                         model_bounds[cont_idx] = (0, int(n_seqs) - 1)
#                     if bounds_ss_cont is None:
#                         lbs[cont_idx] = 0  # in subspace, it should start from 0
#                         ord_hps[param.name] = (0, int(n_seqs))
#                     else:
#                         lbs[cont_idx] = bounds_ss_cont[i][0]  # in subspace, it should start from 0
#                         ord_hps[param.name] = bounds_ss_cont[i]
#                     dims_cont_ord.append(cont_idx)
#                     idx_cont_ord.append(i)
#                 else:
#                     dims_cont_num.append(cont_idx)
#                     idx_cont_num.append(i)

#             if bounds_ss_cat is not None:
#                 self.bounds_ss_cat = [bounds_ss_cat[act_dims_cat_ss] for act_dims_cat_ss in activate_dims_cat_ss]
#             else:
#                 self.bounds_ss_cat = None
#             self.bounds_ss_cont = bounds_ss_cont[activate_dims_cont_ss] if bounds_ss_cont is not None else None

#             if bounds_ss_cont is None:
#                 lbs[dims_cont_num] = 0.0
#                 scales[dims_cont_num] = 1.0
#             else:
#                 lbs[dims_cont_num] = bounds_ss_cont[idx_cont_num, 0]
#                 # rescale numerical hyperparameters to [0., 1.]
#                 scales[dims_cont_num] = 1.0 / (bounds_ss_cont[idx_cont_num, 1] - bounds_ss_cont[idx_cont_num, 0])

#             self.lbs = lbs[activate_dims]
#             self.scales = scales[activate_dims]

#             self.cs_local = ConfigurationSpace()
#             hp_list = []
#             idx_cont = 0
#             idx_cat = 0

#             hps = config_space.get_hyperparameters()

#             for idx in self.activate_dims:
#                 param = hps[idx]
#                 if isinstance(param, CategoricalHyperparameter):
#                     if bounds_ss_cat is None:
#                         hp_new = copy.deepcopy(param)
#                         idx_cat += 1
#                     else:
#                         choices = [param.choices[int(choice_idx)] for choice_idx in bounds_ss_cat[idx_cat]]
#                         # cat_freq_arr = np.array((cats_freq[idx_cat]))
#                         # weights = cat_freq_arr / np.sum(cat_freq_arr)
#                         hp_new = CategoricalHyperparameter(param.name, choices=choices)  # , weights=weights)
#                         idx_cat += 1

#                 elif isinstance(param, OrdinalHyperparameter):
#                     param_seq = ord_hps.get(param.name)
#                     raw_seq = param.sequence
#                     ord_indices = np.arange(*param_seq)
#                     new_seq = [raw_seq[int(round(idx))] for idx in ord_indices]
#                     hp_new = OrdinalHyperparameter(param.name, sequence=new_seq)
#                     idx_cont += 1

#                 elif isinstance(param, Constant):
#                     hp_new = copy.deepcopy(param)
#                 elif isinstance(param, (UniformFloatHyperparameter, UniformIntegerHyperparameter)):
#                     if bounds_ss_cont is None:
#                         hp_new = copy.deepcopy(param)
#                         idx_cont += 1
#                     else:
#                         if isinstance(param, UniformFloatHyperparameter):
#                             lower = param.lower
#                             upper = param.upper
#                             if param.log:
#                                 lower_log = np.log(lower)
#                                 upper_log = np.log(upper)
#                                 hp_new_lower = np.exp((upper_log - lower_log) * bounds_ss_cont[idx_cont][0] +
# lower_log)
#                                 hp_new_upper = np.exp((upper_log - lower_log) * bounds_ss_cont[idx_cont][1] +
# lower_log)
#                                 hp_new = UniformFloatHyperparameter(
#                                     name=param.name,
#                                     lower=max(hp_new_lower, lower),
#                                     upper=min(hp_new_upper, upper),
#                                     log=True,
#                                 )
#                             else:
#                                 hp_new_lower = (upper - lower) * bounds_ss_cont[idx_cont][0] + lower
#                                 hp_new_upper = (upper - lower) * bounds_ss_cont[idx_cont][1] + lower
#                                 hp_new = UniformFloatHyperparameter(
#                                     name=param.name,
#                                     lower=max(hp_new_lower, lower),
#                                     upper=min(hp_new_upper, upper),
#                                     log=False,
#                                 )
#                             idx_cont += 1
#                         elif isinstance(param, UniformIntegerHyperparameter):
#                             lower = param.lower
#                             upper = param.upper
#                             if param.log:
#                                 lower_log = np.log(lower)
#                                 upper_log = np.log(upper)
#                                 hp_new_lower = int(
#                                     math.floor(
#                                         np.exp((upper_log - lower_log) * bounds_ss_cont[idx_cont][0] + lower_log)
#                                     )
#                                 )
#                                 hp_new_upper = int(
#                                     math.ceil(np.exp((upper_log - lower_log) * bounds_ss_cont[idx_cont][1] +
# lower_log))
#                                 )

#                                 hp_new_lower_log = np.log(hp_new_lower)
#                                 hp_new_upper_log = np.log(hp_new_upper)
#                                 new_scale = (upper_log - lower_log) / (hp_new_upper_log - hp_new_lower_log)
#                                 new_lb = (hp_new_lower_log - lower_log) / (hp_new_upper_log - hp_new_lower_log)

#                                 self.scales[idx] = new_scale
#                                 self.lbs[idx] = new_lb

#                                 hp_new = UniformIntegerHyperparameter(
#                                     name=param.name,
#                                     lower=max(hp_new_lower, lower),
#                                     upper=min(hp_new_upper, upper),
#                                     log=True,
#                                 )
#                             else:
#                                 hp_new_lower = int(math.floor((upper - lower) * bounds_ss_cont[idx_cont][0])) + lower
#                                 hp_new_upper = int(math.ceil((upper - lower) * bounds_ss_cont[idx_cont][1])) + lower

#                                 new_scale = (upper - lower) / (hp_new_upper - hp_new_lower)
#                                 new_lb = (hp_new_lower - lower) / (hp_new_upper - hp_new_lower)
#                                 self.scales[idx] = new_scale
#                                 self.lbs[idx] = new_lb

#                                 hp_new = UniformIntegerHyperparameter(
#                                     name=param.name,
#                                     lower=max(hp_new_lower, lower),
#                                     upper=min(hp_new_upper, upper),
#                                     log=False,
#                                 )

#                             idx_cont += 1
#                 else:
#                     raise ValueError(f"Unsupported type of Hyperparameter: {type(param)}")
#                 hp_list.append(hp_new)

#             # We only consider plain hyperparameters
#             self.cs_local.add(hp_list)
#             forbiddens_ss = []
#             forbiddens = config_space.get_forbiddens()
#             for forbidden in forbiddens:
#                 forbiden_ss = self.fit_forbidden_to_ss(cs_local=self.cs_local, forbidden=forbidden)
#                 if forbiden_ss is not None:
#                     forbiddens_ss.append(forbiden_ss)
#             if len(forbiddens_ss) > 0:
#                 self.cs_local.add_forbidden_clauses(forbiddens_ss)

#         model_kwargs = dict(
#             configspace=self.cs_local,
#             # types=[model_types[activate_dim] for activate_dim in activate_dims] if model_types is not None else
# None,
#             # bounds=[model_bounds[activate_dim] for activate_dim in activate_dims] if model_bounds is not None
#             # else None,
#             bounds_cont=np.array([[0, 1.0] for _ in range(len(activate_dims_cont))]),
#             bounds_cat=self.bounds_ss_cat,
#             seed=self.rng.randint(0, 2**20),
#         )

#         if inspect.isclass(model_local):
#             model_local_kwargs_copy = copy.deepcopy(model_local_kwargs)
#             if "kernel_kwargs" in model_local_kwargs_copy:
#                 kernel_kwargs = model_local_kwargs_copy["kernel_kwargs"]
#                 kernel = construct_gp_kernel(kernel_kwargs, activate_dims_cont_ss, activate_dims_cat_ss)
#                 del model_local_kwargs_copy["kernel_kwargs"]
#                 model_local_kwargs_copy["kernel"] = kernel

#             if model_local_kwargs is not None:
#                 model_kwargs.update(model_local_kwargs_copy)

#             all_arguments = inspect.signature(model_local).parameters.keys()
#             if "bounds_cont" not in all_arguments:
#                 del model_kwargs["bounds_cont"]
#             if "bounds_cat" not in all_arguments:
#                 del model_kwargs["bounds_cat"]
#             model = model_local(**model_kwargs)  # type: ignore
#         else:
#             model = model_local

#         self.model = model

#         if inspect.isclass(acq_func_local):
#             acq_func_kwargs = {}
#             if acq_func_local_kwargs is not None:
#                 acq_func_kwargs.update(acq_func_local_kwargs)
#             acquisition_function = acq_func_local(**acq_func_kwargs)  # type: ignore
#         else:
#             acquisition_function = acq_func_local

#         self.acquisition_function = acquisition_function

#         self.incumbent_array = incumbent_array

#         self.model_x = np.empty([0, len(activate_dims)])
#         self.ss_x = np.empty([0, len(activate_dims)])
#         self.model_y = np.empty([0, 1])
#         self.ss_y = np.empty([0, 1])

#         if initial_data is not None:
#             X = initial_data[0]
#             y = initial_data[1]

#             self.add_new_observations(X, y)

#         self.config_origin = "subspace"

#     @staticmethod
#     def fit_forbidden_to_ss(
#         cs_local: ConfigurationSpace, forbidden: AbstractForbiddenComponent
#     ) -> AbstractForbiddenComponent | None:
#         """
#         Fit the forbidden to subspaces. If the target forbidden can be added to subspace, we return a new forbidden
#         with exactly the same type of the input forbidden. Otherwise, None is returned.

#         Parameters
#         ----------
#         cs_local: ConfigurationSpace
#             local configuration space of the subspace
#         forbidden: AbstractForbiddenComponent
#             forbidden to check
#         Returns
#         -------
#         forbidden_ss: AbstractForbiddenComponent | None
#             forbidden in subspaces

#         """
#         if isinstance(forbidden, ForbiddenAndConjunction):
#             forbidden_ss_components = []
#             for forbid in forbidden.components:
#                 # If any of the AndConjunction is not supported by the subspace, we simply ignore them
#                 forbid_ss = LocalSubspace.fit_forbidden_to_ss(cs_local, forbid)
#                 if forbid_ss is None:
#                     return None
#                 forbidden_ss_components.append(forbid_ss)
#             return type(forbidden)(*forbidden_ss_components)
#         else:
#             forbidden_hp_name = forbidden.hyperparameter.name
#             if forbidden_hp_name not in cs_local:
#                 return None
#             hp_ss = cs_local.get_hyperparameter(forbidden_hp_name)

#             def is_value_in_hp(value: Any, hp: Hyperparameter) -> bool:
#                 """Check if the value is in the range of the hp."""
#                 if isinstance(hp, NumericalHyperparameter):
#                     return hp.lower <= value <= hp.upper
#                 elif isinstance(hp, OrdinalHyperparameter):
#                     return value in hp.sequence
#                 elif isinstance(hp, CategoricalHyperparameter):
#                     return value in hp.choices
#                 else:
#                     raise NotImplementedError("Unsupported type of hyperparameter!")

#             if isinstance(forbidden, MultipleValueForbiddenClause):
#                 forbidden_values = forbidden.values
#                 for forbidden_value in forbidden_values:
#                     if not is_value_in_hp(forbidden_value, hp_ss):
#                         return None
#                 return type(forbidden)(hp_ss, forbidden_values)
#             else:
#                 forbidden_value = forbidden.value
#                 if is_value_in_hp(forbidden_value, hp_ss):
#                     return type(forbidden)(hp_ss, forbidden_value)
#             return None

#     def update_model(self, predict_x_best: bool = True, update_incumbent_array: bool = False) -> None:
#         """
#         Update the model and acquisition function parameters

#         Parameters
#         ----------
#         predict_x_best: bool,
#             if the incumbent is acquired by the predicted mean of a surrogate model
#         update_incumbent_array: bool
#             if the incumbent_array of this subspace is replaced with the newly updated incumbent
#         """
#         acq_func_kwargs = {"model": self.model, "num_data": len(self.ss_x)}

#         if predict_x_best:
#             try:
#                 mu, _ = self.model.predict(self.ss_x)
#             except Exception as e:
#                 # Some times it could occur that LGPGA fails to predict the mean value of ss_x because of
#                 # numerical issues
#                 logger.warning(f"Fail to predict ss_x due to {e}")
#                 mu = self.ss_y
#             idx_eta = np.argmin(mu)
#             incumbent_array = self.ss_x[idx_eta]
#             acq_func_kwargs.update({"incumbent_array": incumbent_array, "eta": mu[idx_eta]})
#         else:
#             idx_eta = np.argmin(self.ss_y)
#             incumbent_array = self.ss_x[idx_eta]
#             acq_func_kwargs.update({"incumbent_array": incumbent_array, "eta": self.ss_y[idx_eta]})
#         if update_incumbent_array:
#             if self.incumbent_array is None:
#                 self.incumbent_array = self.ss_x[idx_eta]
#             else:
#                 self.incumbent_array[self.activate_dims] = self.ss_x[idx_eta]

#         self.acquisition_function.update(**acq_func_kwargs)

#     def add_new_observations(self, X: np.ndarray, y: np.ndarray) -> None:
#         """
#         Add new observations to the subspace

#         Parameters
#         ----------
#         X: np.ndarray(N,D),
#             new feature vector of the observations, constructed by the global configuration space
#         y: np.ndarray(N)
#            new performances of the observations
#         Return
#         ----------
#         indices_in_ss:np.ndarray(N)
#             indices of data that included in subspaces
#         """
#         if len(X.shape) == 1:
#             X = X[np.newaxis, :]
#         if len(y.shape) == 1:
#             y = y[:, np.newaxis]

#         X = X[:, self.activate_dims]

#         ss_indices = check_subspace_points(
#             X=X,
#             cont_dims=self.activate_dims_cont,
#             cat_dims=self.activate_dims_cat,
#             bounds_cont=self.bounds_ss_cont,
#             bounds_cat=self.bounds_ss_cat,
#         )

#         X = self.normalize_input(X=X)

#         self.model_x = np.vstack([self.model_x, X])
#         self.model_y = np.vstack([self.model_y, y])

#         self.ss_x = np.vstack([self.ss_x, X[ss_indices]])
#         self.ss_y = np.vstack([self.ss_y, y[ss_indices]])

#     def update_incumbent_array(self, new_incumbent: np.ndarray) -> None:
#         """
#         Update a new incumbent array. The array is generated from the global configuration

#         Parameters
#         ----------
#         new_incumbent: np.ndarray(D)
#             new incumbent, which correspondences to the global configuration
#         """
#         self.incumbent_array = self.normalize_input(X=new_incumbent)

#     def generate_challengers(self, **optimizer_kwargs: Any) -> Iterator:
#         """
#         Generate a list of challengers that will be transformed into the global configuration space

#         Parameters
#         ----------
#         optimizer_kwargs: Any
#             additional configurations passed to 'self._generate_challengers'

#         Returns
#         -------
#             A list of challengers in the global configuration space

#         """
#         challengers = self._generate_challengers(**optimizer_kwargs)
#         return ChallengerListLocal(
#             cs_local=self.cs_local,
#             cs_global=self.cs_global,
#             challengers=challengers,
#             config_origin=self.config_origin,
#             incumbent_array=self.incumbent_array,
#         )

#     @abstractmethod
#     def _generate_challengers(self, **optimizer_kwargs: Any) -> List[Tuple[float, Configuration]]:
#         """Generate new challengers list for this subspace"""
#         raise NotImplementedError

#     def normalize_input(self, X: np.ndarray) -> np.ndarray:
#         """
#         Normalize X to fit the local configuration space

#         Parameters
#         ----------
#         X: np.ndarray(N,D)
#             input X, configurations arrays
#         Returns
#         -------
#         X_normalized: np.ndarray(N,D)
#             normalized input X
#         """
#         if not self.new_config:
#             return X

#         if len(X.shape) == 1:
#             X = X[np.newaxis, :]

#         # normalize X
#         X_normalized = (X - self.lbs) * self.scales
#         if self.bounds_ss_cat is not None:
#             # normalize categorical function, for instance, if bounds_subspace[i] is a categorical bound contains
#             # elements [1, 3, 5], then we map 1->0, 3->1, 5->2
#             for cat_idx, cat_bound in zip(self.activate_dims_cat, self.bounds_ss_cat):
#                 X_i = X_normalized[:, cat_idx]
#                 cond_list = [X_i == cat for cat in cat_bound]
#                 choice_list = np.arange(len(cat_bound))
#                 X_i = np.select(cond_list, choice_list)
#                 X_normalized[:, cat_idx] = X_i

#         return X_normalized


# class ChallengerListLocal(Iterator):
#     def __init__(
#         self,
#         cs_local: ConfigurationSpace,
#         cs_global: ConfigurationSpace,
#         challengers: List[Tuple[float, Configuration]],
#         config_origin: str,
#         incumbent_array: np.ndarray | None = None,
#     ):
#         """
#         A Challenger list to convert the configuration from the local configuration space to the global configuration
#          space

#         Parameters
#         ----------
#         cs_local: ConfigurationSpace
#             local configuration space
#         cs_global: ConfigurationSpace
#             global configuration space
#         challengers: List[Tuple[float, Configuration]],
#             challenger lists
#         config_origin: str
#             configuration origin
#         incumbent_array: np.ndarray | None = None,
#             global incumbent array, used when cs_local and cs_global have different number of dimensions and we need
# to
#             supplement the missing values.
#         """
#         self.cs_local = cs_local
#         self.challengers = challengers
#         self.cs_global = cs_global
#         self._index = 0
#         self.config_origin = config_origin
#         # In case cs_in and cs_out have different dimensions
#         self.expand_dims = len(cs_global.get_hyperparameters()) != len(cs_local.get_hyperparameters())
#         self.incumbent_array = incumbent_array

#         if self.expand_dims and self.incumbent_array is None:
#             raise ValueError(
#                 "Incumbent array must be provided if the global configuration space has more "
#                 "hyperparameters then the local configuration space"
#             )

#     def __next__(self) -> Configuration:
#         if self.challengers is not None and self._index == len(self.challengers):
#             raise StopIteration
#         challenger = self.challengers[self._index][1]
#         self._index += 1
#         value = challenger.get_dictionary()
#         if self.expand_dims:
#             incumbent_array = Configuration(
#                 configuration_space=self.cs_global, vector=self.incumbent_array
#             ).get_dictionary()
#             # we replace the cooresponding value in incumbent array with the value suggested by our optimizer
#             for k in value.keys():
#                 incumbent_array[k] = value[k]
#             config = Configuration(configuration_space=self.cs_global, values=incumbent_array)
#         else:
#             config = Configuration(configuration_space=self.cs_global, values=value)
#         if self.config_origin is not None:
#             config.origin = self.config_origin
#         else:
#             config.origin = challenger.origin
#         return config

#     def __len__(self) -> int:
#         if self.challengers is None:
#             self.challengers = []
#         return len(self.challengers) - self._index



from __future__ import annotations

from typing import Callable, Iterator

from ConfigSpace import Configuration, ConfigurationSpace

from smac.random_design import ProbabilityRandomDesign
from smac.random_design.abstract_random_design import AbstractRandomDesign


class ChallengerList(Iterator):
    """Helper class to interleave random configurations in a list of challengers.

    Provides an iterator which returns a random configuration in each second
    iteration. Reduces time necessary to generate a list of new challengers
    as one does not need to sample several hundreds of random configurations
    in each iteration which are never looked at.

    Parameters
    ----------
    configspace : ConfigurationSpace
    challenger_callback : Callable
        Callback function which returns a list of challengers (without interleaved random configurations), must a be a
        python closure.
    random_design : AbstractRandomDesign | None, defaults to ModulusRandomDesign(modulus=2.0)
        Which random design should be used.
    """

    def __init__(
        self,
        configspace: ConfigurationSpace,
        challenger_callback: Callable,
        random_design: AbstractRandomDesign | None = ProbabilityRandomDesign(seed=0, probability=0.08447232371720552),
    ):
        self._challengers_callback = challenger_callback
        self._challengers: list[Configuration] | None = None
        self._configspace = configspace
        self._index = 0
        self._iteration = 1  # 1-based to prevent from starting with a random configuration
        self._random_design = random_design

    def __next__(self) -> Configuration:
        # If we already returned the required number of challengers
        if self._challengers is not None and self._index == len(self._challengers):
            raise StopIteration
        # If we do not want to have random configs, we just yield the next challenger
        elif self._random_design is None:
            if self._challengers is None:
                self._challengers = self._challengers_callback()

            config = self._challengers[self._index]
            self._index += 1

            return config
        # If we want to interleave challengers with random configs, sample one
        else:
            if self._random_design.check(self._iteration):
                config = self._configspace.sample_configuration()
                config.origin = "Random Search"
            else:
                if self._challengers is None:
                    self._challengers = self._challengers_callback()

                config = self._challengers[self._index]
                self._index += 1
            self._iteration += 1

            return config

    def __len__(self) -> int:
        if self._challengers is None:
            self._challengers = self._challengers_callback()

        return len(self._challengers) - self._index



from __future__ import annotations

from ConfigSpace import Configuration

from smac.initial_design.abstract_initial_design import AbstractInitialDesign

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class RandomInitialDesign(AbstractInitialDesign):
    """Initial design that evaluates random configurations."""

    def _select_configurations(self) -> list[Configuration]:
        if self._n_configs == 1:
            configs = [self._configspace.sample_configuration()]
        else:
            configs = self._configspace.sample_configuration(size=self._n_configs)
        for config in configs:
            config.origin = "Initial Design: Random"
        return configs



from __future__ import annotations

import itertools

import numpy as np
from ConfigSpace.configuration_space import Configuration
from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    Constant,
    NumericalHyperparameter,
    OrdinalHyperparameter,
)
from ConfigSpace.util import deactivate_inactive_hyperparameters

from smac.initial_design.abstract_initial_design import AbstractInitialDesign

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


class FactorialInitialDesign(AbstractInitialDesign):
    """Factorial initial design to select corner and middle configurations."""

    def _select_configurations(self) -> list[Configuration]:
        params = list(self._configspace.values())

        values = []
        mid = []
        for param in params:
            if isinstance(param, Constant):
                v = [param.value]
                mid.append(param.value)
            elif isinstance(param, NumericalHyperparameter):
                v = [param.lower, param.upper]
                mid.append(np.average([param.lower, param.upper]))
            elif isinstance(param, CategoricalHyperparameter):
                v = list(param.choices)
                mid.append(param.choices[0])
            elif isinstance(param, OrdinalHyperparameter):
                v = [param.sequence[0], param.sequence[-1]]
                length = len(param.sequence)
                mid.append(param.sequence[int(length / 2)])

            values.append(v)

        factorial_design = itertools.product(*values)

        configs = [self._configspace.get_default_configuration()]
        # add middle point in space
        conf_dict = dict([(p.name, v) for p, v in zip(params, mid)])
        middle_conf = deactivate_inactive_hyperparameters(conf_dict, self._configspace)
        configs.append(middle_conf)

        # Add corner points
        for design in factorial_design:
            conf_dict = dict([(p.name, v) for p, v in zip(params, design)])
            conf = deactivate_inactive_hyperparameters(conf_dict, self._configspace)
            conf.origin = "Initial Design: Factorial"
            configs.append(conf)

        return configs



from __future__ import annotations

from typing import Any

from dataclasses import dataclass, field

from ConfigSpace import Configuration

from smac.runhistory.enumerations import StatusType

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"


@dataclass(frozen=True)
class InstanceSeedKey:
    """Key for instance and seed.

    Parameters
    ----------
    instance : str | None, defaults to None
    seed : int | None, defaults to None
    """

    instance: str | None = None
    seed: int | None = None

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, InstanceSeedKey):
            if self.instance == other.instance and self.seed == other.seed:
                return True

        return False


@dataclass(frozen=True)
class InstanceSeedBudgetKey:
    """Key for instance, seed and budget.

    Parameters
    ----------
    instance : str | None, defaults to None
    seed : int | None, defaults to None
    budget : float | None, defaults to None
    """

    instance: str | None = None
    seed: int | None = None
    budget: float | None = None

    def __lt__(self, other: InstanceSeedBudgetKey) -> bool:
        if self.budget is not None and other.budget is not None:
            return self.budget < other.budget

        if self.instance is not None and other.instance is not None:
            return self.instance < other.instance

        if self.seed is not None and other.seed is not None:
            return self.seed < other.seed

        raise RuntimeError("Could not compare InstanceSeedBudgetKey.")

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, InstanceSeedBudgetKey):
            if self.instance == other.instance and self.seed == other.seed and self.budget == other.budget:
                return True

        return False

    def get_instance_seed_key(self) -> InstanceSeedKey:
        """Returns the instance-seed key. The budget is omitted."""
        return InstanceSeedKey(instance=self.instance, seed=self.seed)


@dataclass(frozen=True)
class TrialKey:
    """Key of a trial.

    Parameters
    ----------
    config_id : int
    instance : str | None, defaults to None
    seed : int | None, defaults to None
    budget : float | None, defaults to None
    """

    config_id: int
    instance: str | None = None
    seed: int | None = None
    budget: float | None = None


@dataclass(frozen=True)
class TrialValue:
    """Values of a trial.

    Parameters
    ----------
    cost : float | list[float]
    time : float, defaults to 0.0
    cpu_time : float, defaults to 0.0
        Describes the amount of time the trial spend on hardware.
    status : StatusType, defaults to StatusType.SUCCESS
    starttime : float, defaults to 0.0
    endtime : float, defaults to 0.0
    additional_info : dict[str, Any], defaults to {}
    """

    cost: float | list[float]
    time: float = 0.0
    cpu_time: float = 0.0
    status: StatusType = StatusType.SUCCESS
    starttime: float = 0.0
    endtime: float = 0.0
    additional_info: dict[str, Any] = field(default_factory=dict)


@dataclass(frozen=True)
class TrialInfo:
    """Information about a trial.

    Parameters
    ----------
    config : Configuration
    instance : str | None, defaults to None
    seed : int | None, defaults to None
    budget : float | None, defaults to None
    """

    config: Configuration
    instance: str | None = None
    seed: int | None = None
    budget: float | None = None

    def get_instance_seed_key(self) -> InstanceSeedKey:
        """Instantiates and returns an InstanceSeedKey object"""
        return InstanceSeedKey(instance=self.instance, seed=self.seed)

    def get_instance_seed_budget_key(self) -> InstanceSeedBudgetKey:
        """Instantiates and returns an InstanceSeedBudgetKey object."""
        return InstanceSeedBudgetKey(instance=self.instance, seed=self.seed, budget=self.budget)


@dataclass
class TrajectoryItem:
    """Item of a trajectory.

    Parameters
    ----------
    config_ids : list[int]
        Configuration ids of the current incumbents.
    costs : list[float | list[float]]
        Costs of the current incumbents. In case of multi-objective, this is a list of lists.
    trial : int
        How many trials have been evaluated so far.
    walltime : float
        How much walltime has been used so far.
    """

    config_ids: list[int]
    costs: list[float | list[float]]
    trial: int
    walltime: float



from __future__ import annotations

from typing import Any
from src.tasks.task import Task
from src.utils.exceptions import NotSupportedError
from src.wrappers.wrapper import Wrapper


class Version14(Wrapper):
    supported_versions: list[str] = ["1.4.0"]

    def __init__(self, task: Task, seed: int) -> None:
        super().__init__(task, seed)

        from smac.facade.smac_ac_facade import SMAC4AC

        self._smac: SMAC4AC | None = None

    def run(self) -> None:
        from smac.facade.smac_bb_facade import SMAC4BB
        from smac.facade.smac_hpo_facade import SMAC4HPO
        from smac.facade.smac_mf_facade import SMAC4MF
        from smac.facade.smac_ac_facade import SMAC4AC
        from smac.scenario.scenario import Scenario

        if self.task.n_workers > 1 and self.task.optimization_type != "mf":
            raise NotSupportedError("SMAC 1.4 does not support parallel execution natively.")

        if self.task.retrain_after > 1:
            raise NotSupportedError("SMAC 1.4 does not support ``retrain_after`.")

        # Get instances
        if self.task.use_instances:
            dataset = self.task.model.dataset
            assert dataset is not None
            all_feat = dataset.get_instance_features()

            instances = []
            instance_features = {}

            # We map the instance to its index (idk why they implemented it this way)
            for i, instance in enumerate(dataset.get_instances()):
                instances.append([instance])
                instance_features[instance] = all_feat[instance]
        else:
            instances = [[None]]
            instance_features = {}

        multi_objectives: str | list[str]
        if len(self.task.objectives) > 1:
            multi_objectives = self.task.objectives
        elif len(self.task.objectives) == 1:
            multi_objectives = self.task.objectives[0]

        scenario = Scenario(
            {
                "run_obj": "quality",  # we optimize quality (alternatively runtime)
                "runcount-limit": self.task.n_trials,  # max. number of function evaluations
                "cs": self.task.model.configspace,  # configuration space
                "wallclock_limit": self.task.walltime_limit,
                "deterministic": self.task.deterministic,
                "min_budget": self.task.min_budget,
                "max_budget": self.task.max_budget,
                "minR": 1,
                "maxR": self.task.max_config_calls,
                "instances": instances,
                "feature_dict": instance_features,
                "multi_objectives": multi_objectives,
                "seed": self.seed,
            }
        )

        intensifier_kwargs: dict[Any, Any] = {}
        facade_kwargs: dict[Any, Any] = {}

        # Create facade
        if self.task.optimization_type == "bb":
            facade_object = SMAC4BB
            intensifier_kwargs["maxR"] = self.task.max_config_calls

        elif self.task.optimization_type == "hpo":
            facade_object = SMAC4HPO
            intensifier_kwargs["maxR"] = self.task.max_config_calls

        elif self.task.optimization_type == "mf":
            facade_object = SMAC4MF

            n_seeds = self.task.n_seeds
            assert n_seeds is not None
            intensifier_kwargs["n_seeds"] = n_seeds
            intensifier_kwargs["initial_budget"] = self.task.min_budget
            intensifier_kwargs["max_budget"] = self.task.max_budget

            inc_selection = self.task.incumbent_selection
            if inc_selection == "highest_observed_budget":
                inc_selection = "highest_budget"

            intensifier_kwargs["incumbent_selection"] = inc_selection
            facade_kwargs["n_jobs"] = self.task.n_workers
        elif self.task.optimization_type == "ac":
            facade_object = SMAC4AC
            intensifier_kwargs["maxR"] = self.task.max_config_calls

        else:
            raise RuntimeError("Unknown optimization type.")

        if self.task.intensifier is None:
            intensifier = None
        else:
            if self.task.intensifier == "successive_halving":
                from smac.intensification.successive_halving import SuccessiveHalving

                intensifier = SuccessiveHalving
            else:
                raise RuntimeError("Unsupported intensifier.")

        smac = facade_object(
            scenario=scenario,
            tae_runner=self.task.model.train,
            intensifier=intensifier,
            intensifier_kwargs=intensifier_kwargs,
            **facade_kwargs,
        )
        smac.optimize()

        self._smac = smac

    def get_trajectory(self, sort_by: str = "trials") -> tuple[list[float], list[float]]:

        if len(self.task.objectives) > 1:
            raise NotSupportedError

        assert self._smac is not None
        trajectory = self._smac.trajectory
        X: list[int | float] = []
        Y: list[float] = []

        for (cost, _, _, trial, _, walltime, _) in trajectory:
            if cost > 1e6:
                continue

            if sort_by == "trials":
                X.append(trial)
            elif sort_by == "walltime":
                X.append(walltime)
            else:
                raise RuntimeError("Unknown sort_by.")

            Y.append(cost)

        return X, Y



from __future__ import annotations

from typing import Any, Iterator

import math
from collections import defaultdict

import numpy as np
from ConfigSpace import Configuration

from smac.constants import MAXINT
from smac.intensifier.abstract_intensifier import AbstractIntensifier
from smac.runhistory import TrialInfo
from smac.runhistory.dataclasses import InstanceSeedBudgetKey
from smac.runhistory.errors import NotEvaluatedError
from smac.scenario import Scenario
from smac.utils.configspace import get_config_hash
from smac.utils.data_structures import batch
from smac.utils.logging import get_logger
from smac.utils.pareto_front import calculate_pareto_front, sort_by_crowding_distance

__copyright__ = "Copyright 2025, Leibniz University Hanover, Institute of AI"
__license__ = "3-clause BSD"

logger = get_logger(__name__)


class SuccessiveHalving(AbstractIntensifier):
    """
    Implementation of Succesive Halving supporting multi-fidelity, multi-objective, and multi-processing.
    Internally, a tracker keeps track of configurations and their bracket and stage.

    The behaviour of this intensifier is as follows:

    - First, adds configurations from the runhistory to the tracker. The first stage is always filled-up. For example,
      the user provided 4 configs with the tell-method but the first stage requires 8 configs: 4 new configs are
      sampled and added together with the provided configs as a group to the tracker.
    - While loop:

      - If a trial in the tracker has not been yielded yet, yield it.
      - If we are running out of trials, we simply add a new batch of configurations to the first stage.

    Note
    ----
    The implementation natively supports brackets from Hyperband. However, in the case of Successive Halving,
    only one bracket is used.

    Parameters
    ----------
    eta : int, defaults to 3
        Input that controls the proportion of configurations discarded in each round of Successive Halving.
    n_seeds : int, defaults to 1
        How many seeds to use for each instance.
    instance_seed_order : str, defaults to "shuffle_once"
        How to order the instance-seed pairs. Can be set to:

        - `None`: No shuffling at all and use the instance-seed order provided by the user.
        - `shuffle_once`: Shuffle the instance-seed keys once and use the same order across all runs.
        - `shuffle`: Shuffles the instance-seed keys for each bracket individually.
    incumbent_selection : str, defaults to "highest_observed_budget"
        How to select the incumbent when using budgets. Can be set to:

        - `any_budget`: Incumbent is the best on any budget i.e., best performance regardless of budget.
        - `highest_observed_budget`: Incumbent is the best in the highest budget run so far.
        - `highest_budget`: Incumbent is selected only based on the highest budget.
    max_incumbents : int, defaults to 10
        How many incumbents to keep track of in the case of multi-objective.
    seed : int, defaults to None
        Internal seed used for random events like shuffle seeds.
    """

    def __init__(
        self,
        scenario: Scenario,
        eta: int = 3,
        n_seeds: int = 1,
        instance_seed_order: str | None = "shuffle_once",
        max_incumbents: int = 10,
        incumbent_selection: str = "highest_observed_budget",
        seed: int | None = None,
    ):
        super().__init__(
            scenario=scenario,
            n_seeds=n_seeds,
            max_incumbents=max_incumbents,
            seed=seed,
        )

        self._eta = eta
        self._instance_seed_order = instance_seed_order
        self._incumbent_selection = incumbent_selection
        self._highest_observed_budget_only = False if incumbent_selection == "any_budget" else True

        # Global variables derived from scenario
        self._min_budget = self._scenario.min_budget
        self._max_budget = self._scenario.max_budget

    @property
    def meta(self) -> dict[str, Any]:  # noqa: D102
        meta = super().meta
        meta.update(
            {
                "eta": self._eta,
                "instance_seed_order": self._instance_seed_order,
                "incumbent_selection": self._incumbent_selection,
            }
        )

        return meta

    def reset(self) -> None:
        """Reset the internal variables of the intensifier including the tracker."""
        super().reset()

        # States
        # dict[tuple[bracket, stage], list[tuple[seed to shuffle instance-seed keys, list[config_id]]]
        self._tracker: dict[tuple[int, int], list[tuple[int | None, list[Configuration]]]] = defaultdict(list)

    def __post_init__(self) -> None:
        """Post initialization steps after the runhistory has been set."""
        super().__post_init__()

        # We generate our instance seed pairs once
        is_keys = self.get_instance_seed_keys_of_interest()

        # Budgets, followed by lots of sanity-checking
        eta = self._eta
        min_budget = self._min_budget
        max_budget = self._max_budget

        if max_budget is not None and min_budget is not None and max_budget < min_budget:
            raise ValueError("Max budget has to be larger than min budget.")

        if self.uses_instances:
            if isinstance(min_budget, float) or isinstance(max_budget, float):
                raise ValueError("Successive Halving requires integer budgets when using instances.")

            min_budget = min_budget if min_budget is not None else 1
            max_budget = max_budget if max_budget is not None else len(is_keys)

            if max_budget > len(is_keys):
                raise ValueError(
                    f"Max budget of {max_budget} can not be greater than the number of instance-seed "
                    f"keys ({len(is_keys)})."
                )

            if max_budget < len(is_keys):
                logger.warning(
                    f"Max budget {max_budget} does not include all instance seed  " f"pairs ({len(is_keys)})."
                )
        else:
            if min_budget is None or max_budget is None:
                raise ValueError(
                    "Successive Halving requires the parameters min_budget and max_budget defined in the scenario."
                )

            if len(is_keys) != 1:
                raise ValueError("Successive Halving supports only one seed when using budgets.")

        if min_budget is None or min_budget <= 0:
            raise ValueError("Min budget has to be larger than 0.")

        budget_type = "INSTANCES" if self.uses_instances else "BUDGETS"
        logger.info(
            f"Successive Halving uses budget type {budget_type} with eta {eta}, "
            f"min budget {min_budget}, and max budget {max_budget}."
        )

        # Pre-computing Successive Halving variables
        max_iter = self._get_max_iterations(eta, max_budget, min_budget)
        budgets, n_configs = self._compute_configs_and_budgets_for_stages(eta, max_budget, max_iter)

        # Global variables
        self._min_budget = min_budget
        self._max_budget = max_budget

        # Stage variables, depending on the bracket (0 is the bracket here since SH only has one bracket)
        self._max_iterations: dict[int, int] = {0: max_iter + 1}
        self._n_configs_in_stage: dict[int, list] = {0: n_configs}
        self._budgets_in_stage: dict[int, list] = {0: budgets}

    @staticmethod
    def _get_max_iterations(eta: int, max_budget: float | int, min_budget: float | int) -> int:
        return int(np.floor(np.log(max_budget / min_budget) / np.log(eta)))

    @staticmethod
    def _compute_configs_and_budgets_for_stages(
        eta: int, max_budget: float | int, max_iter: int, s_max: int | None = None
    ) -> tuple[list[int], list[int]]:
        if s_max is None:
            s_max = max_iter

        n_initial_challengers = math.ceil((eta**max_iter) * (s_max + 1) / (max_iter + 1))

        # How many configs in each stage
        lin_space = -np.linspace(0, max_iter, max_iter + 1)
        n_configs_ = np.floor(n_initial_challengers * np.power(eta, lin_space))
        n_configs = np.array(np.round(n_configs_), dtype=int).tolist()

        # How many budgets in each stage
        lin_space = -np.linspace(max_iter, 0, max_iter + 1)
        budgets = (max_budget * np.power(eta, lin_space)).tolist()

        return budgets, n_configs

    def get_state(self) -> dict[str, Any]:  # noqa: D102
        # Replace config by dict
        tracker: dict[str, list[tuple[int | None, list[dict]]]] = defaultdict(list)
        for key in list(self._tracker.keys()):
            for seed, configs in self._tracker[key]:
                # We have to make key serializable
                new_key = f"{key[0]},{key[1]}"
                tracker[new_key].append((seed, [dict(config) for config in configs]))

        return {"tracker": tracker}

    def set_state(self, state: dict[str, Any]) -> None:  # noqa: D102
        self._tracker = defaultdict(list)

        tracker = state["tracker"]
        for old_key in list(tracker.keys()):
            keys = [k for k in old_key.split(",")]
            new_key = (int(keys[0]), int(keys[1]))
            for seed, config_dicts in tracker[old_key]:
                seed = None if seed is None else int(seed)
                self._tracker[new_key].append(
                    (
                        seed,
                        [Configuration(self._scenario.configspace, config_dict) for config_dict in config_dicts],
                    )
                )

    @property
    def uses_seeds(self) -> bool:  # noqa: D102
        return True

    @property
    def uses_budgets(self) -> bool:  # noqa: D102
        if self._scenario.instances is None:
            return True

        return False

    @property
    def uses_instances(self) -> bool:  # noqa: D102
        if self._scenario.instances is None:
            return False

        return True

    def print_tracker(self) -> None:
        """Prints the number of configurations in each bracket/stage."""
        messages = []
        for (bracket, stage), others in self._tracker.items():
            counter = 0
            for _, config_ids in others:
                counter += len(config_ids)

            if counter > 0:
                messages.append(f"--- Bracket {bracket} / Stage {stage}: {counter} configs")

        if len(messages) > 0:
            logger.debug(f"{self.__class__.__name__} statistics:")

        for message in messages:
            logger.debug(message)

    def get_trials_of_interest(
        self,
        config: Configuration,
        *,
        validate: bool = False,
        seed: int | None = None,
    ) -> list[TrialInfo]:  # noqa: D102
        is_keys = self.get_instance_seed_keys_of_interest(validate=validate, seed=seed)
        budget = None

        # When we use budgets, we always evaluated on the highest budget only
        if self.uses_budgets:
            budget = self._max_budget

        trials = []
        for key in is_keys:
            trials.append(TrialInfo(config=config, instance=key.instance, seed=key.seed, budget=budget))

        return trials

    def get_instance_seed_budget_keys(
        self, config: Configuration, compare: bool = False
    ) -> list[InstanceSeedBudgetKey]:
        """Returns the instance-seed-budget keys for a given configuration. This method supports ``highest_budget``,
        which only returns the instance-seed-budget keys for the highest budget (if specified). In this case, the
        incumbents in ``update_incumbents`` are only changed if the costs on the highest budget are lower.

        Parameters
        ----------
        config: Configuration
            The Configuration to be queried
        compare : bool, defaults to False
            Get rid of the budget information for comparing if the configuration was evaluated on the same
            instance-seed keys.
        """
        isb_keys = self.runhistory.get_instance_seed_budget_keys(
            config, highest_observed_budget_only=self._highest_observed_budget_only
        )

        # If incumbent should only be changed on the highest budget, we have to kick out all budgets below the highest
        if self.uses_budgets and self._incumbent_selection == "highest_budget":
            isb_keys = [key for key in isb_keys if key.budget == self._max_budget]

        if compare:
            # Get rid of duplicates
            isb_keys = list(
                set([InstanceSeedBudgetKey(instance=key.instance, seed=key.seed, budget=None) for key in isb_keys])
            )

        return isb_keys

    def __iter__(self) -> Iterator[TrialInfo]:  # noqa: D102
        self.__post_init__()

        # Log brackets/stages
        logger.info("Number of configs in stage:")
        for bracket, n in self._n_configs_in_stage.items():
            logger.info(f"--- Bracket {bracket}: {n}")

        logger.info("Budgets in stage:")
        for bracket, budgets in self._budgets_in_stage.items():
            logger.info(f"--- Bracket {bracket}: {budgets}")

        rh = self.runhistory

        # We have to add already existing trials from the runhistory
        # Idea: We simply add existing configs to the tracker (first stage) but assign a random instance shuffle seed.
        # In the best case, trials (added from the users) are included in the seed and it has not re-computed again.
        # Note: If the intensifier was restored, we don't want to go in here
        if len(self._tracker) == 0:
            bracket = 0
            stage = 0

            # Print ignored budgets
            ignored_budgets = []
            for k in rh.keys():
                if k.budget not in self._budgets_in_stage[0] and k.budget not in ignored_budgets:
                    ignored_budgets.append(k.budget)

            if len(ignored_budgets) > 0:
                logger.warning(
                    f"Trials with budgets {ignored_budgets} will been ignored. Consider adding trials with budgets "
                    f"{self._budgets_in_stage[0]}."
                )

            # We batch the configs because we need n_configs in each iteration
            # If we don't have n_configs, we sample new ones
            # We take the number of configs from the first bracket and the first stage
            n_configs = self._n_configs_in_stage[bracket][stage]
            for configs in batch(rh.get_configs(), n_configs):
                n_rh_configs = len(configs)

                if len(configs) < n_configs:
                    try:
                        config = next(self.config_generator)
                        configs.append(config)
                    except StopIteration:
                        # We stop if we don't find any configuration anymore
                        return

                seed = self._get_next_order_seed()
                self._tracker[(bracket, stage)].append((seed, configs))
                logger.info(
                    f"Added {n_rh_configs} configs from runhistory and {n_configs - n_rh_configs} new configs to "
                    f"Successive Halving's first bracket and first stage with order seed {seed}."
                )

        while True:
            # If we don't yield trials anymore, we have to update
            # Otherwise, we can just keep yielding trials from the tracker
            update = False

            # We iterate over the tracker to do two things:
            # 1) Yield trials of configs that are not yet evaluated/running
            # 2) Update tracker and move better configs to the next stage
            # We start in reverse order to complete higher stages first
            logger.debug("Updating tracker:")

            # TODO: Process stages ascending or descending?
            for bracket, stage in list(self._tracker.keys()):
                pairs = self._tracker[(bracket, stage)].copy()
                for seed, configs in pairs:
                    isb_keys = self._get_instance_seed_budget_keys_by_stage(bracket=bracket, stage=stage, seed=seed)

                    # We iterate over the configs and yield trials which are not running/evaluated yet
                    for config in configs:
                        config_hash = get_config_hash(config)
                        trials = self._get_next_trials(config, from_keys=isb_keys)
                        logger.debug(
                            f"--- Yielding {len(trials)}/{len(isb_keys)} for config {config_hash} in "
                            f"stage {stage} with seed {seed}..."
                        )

                        for trial in trials:
                            yield trial
                            update = True

                    # If all configs were evaluated on ``n_configs_required``, we finally can compare
                    try:
                        successful_configs = self._get_best_configs(configs, bracket, stage, isb_keys)
                    except NotEvaluatedError:
                        # We can't compare anything, so we just continue with the next pairs
                        logger.debug("--- Could not compare configs because not all trials have been evaluated yet.")
                        continue

                    # Update tracker
                    # Remove current shuffle index / config pair
                    self._tracker[(bracket, stage)].remove((seed, configs))

                    # Add successful to the next stage
                    if stage < self._max_iterations[bracket] - 1:
                        config_ids = [rh.get_config_id(config) for config in successful_configs]
                        self._tracker[(bracket, stage + 1)].append((seed, successful_configs))

                        logger.debug(
                            f"--- Promoted {len(config_ids)} configs from stage {stage} to stage {stage + 1} in "
                            f"bracket {bracket}."
                        )
                    else:
                        logger.debug(
                            f"--- Removed {len(successful_configs)} configs to last stage in bracket {bracket}."
                        )

                    # Log how many configs are in each stage
                    self.print_tracker()

            # Since we yielded something before, we want to go back as long as we do not find any trials anymore
            if update:
                continue

            # TODO: Aggressive progressing without knowing how well trials performed
            # Idea: Don't add constantly new batches (see ASHA)

            # If we are running out of trials, we want to add configs to the first stage
            # We simply add as many configs to the stage as required (_n_configs_in_stage[0])
            configs = []
            next_bracket = self._get_next_bracket()
            for _ in range(self._n_configs_in_stage[next_bracket][0]):
                try:
                    config = next(self.config_generator)
                    configs.append(config)
                except StopIteration:
                    # We stop if we don't find any configuration anymore
                    logger.warning(
                        "If you assume your configspace was not yet exhausted, try to "
                        "increase the number of retries in the config selector."
                    )
                    return

            # We keep track of the seed so we always evaluate on the same instances
            next_seed = self._get_next_order_seed()
            self._tracker[(next_bracket, 0)].append((next_seed, configs))
            logger.debug(
                f"Added {len(configs)} new configs to bracket {next_bracket} stage 0 with shuffle seed {next_seed}."
            )

    def _get_instance_seed_budget_keys_by_stage(
        self,
        bracket: int,
        stage: int,
        seed: int | None = None,
    ) -> list[InstanceSeedBudgetKey]:
        """Returns all instance-seed-budget keys (isb keys) for the given stage. Each stage
        is associated with a budget (N). Two possible options:

        1) Instance based: We return N isb keys. If a seed is specified, we shuffle the keys before
        returning the first N instances. The budget is set to None here.
        2) Budget based: We return one isb only but the budget is set to N.
        """
        budget: float | int | None = None
        is_keys = self.get_instance_seed_keys_of_interest()

        # We have to differentiate between budgets and instances based here
        # If we are budget based, we always have one instance seed pair only
        # If we are in the instance setting, we have to return a specific number of instance seed pairs

        if self.uses_instances:
            # Shuffle instance seed pairs group-based
            if seed is not None:
                is_keys = self._reorder_instance_seed_keys(is_keys, seed=seed)

            # We only return the first N instances
            N = int(self._budgets_in_stage[bracket][stage])
            is_keys = is_keys[:N]
        else:
            assert len(is_keys) == 1

            # The stage defines which budget should be used (in real-valued setting)
            # No shuffle is needed here because we only have on instance seed pair
            budget = self._budgets_in_stage[bracket][stage]

        isbk = []
        for isk in is_keys:
            isbk.append(InstanceSeedBudgetKey(instance=isk.instance, seed=isk.seed, budget=budget))

        return isbk

    def _get_next_trials(
        self,
        config: Configuration,
        from_keys: list[InstanceSeedBudgetKey],
    ) -> list[TrialInfo]:
        """Returns trials for a given config from a list of instances (instance-seed-budget keys). The returned trials
        have not run or evaluated yet.
        """
        rh = self.runhistory
        evaluated_trials = rh.get_trials(config, highest_observed_budget_only=False)
        running_trials = rh.get_running_trials(config)

        next_trials: list[TrialInfo] = []
        for instance in from_keys:
            trial = TrialInfo(config=config, instance=instance.instance, seed=instance.seed, budget=instance.budget)

            if trial in evaluated_trials or trial in running_trials:
                continue

            next_trials.append(trial)

        return next_trials

    def _get_best_configs(
        self,
        configs: list[Configuration],
        bracket: int,
        stage: int,
        from_keys: list[InstanceSeedBudgetKey],
    ) -> list[Configuration]:
        """Returns the best configurations. The number of configurations is depending on the stage. Raises
        ``NotEvaluatedError`` if not all trials have been evaluated.
        """
        try:
            n_configs = self._n_configs_in_stage[bracket][stage + 1]
        except IndexError:
            return []

        rh = self.runhistory
        configs = configs.copy()

        for config in configs:
            isb_keys = rh.get_instance_seed_budget_keys(config)
            if not all(isb_key in isb_keys for isb_key in from_keys):
                raise NotEvaluatedError

        selected_configs: list[Configuration] = []
        while len(selected_configs) < n_configs:
            # We calculate the pareto front for the given configs
            # We use the same isb keys for all the configs
            all_keys = [from_keys for _ in configs]
            incumbents = calculate_pareto_front(rh, configs, all_keys)

            # Idea: We recursively calculate the pareto front in every iteration
            for incumbent in incumbents:
                configs.remove(incumbent)
                selected_configs.append(incumbent)

        # If we have more selected configs, we remove the ones with the smallest crowding distance
        if len(selected_configs) > n_configs:
            all_keys = [from_keys for _ in selected_configs]
            selected_configs = sort_by_crowding_distance(rh, selected_configs, all_keys)[:n_configs]
            logger.debug("Found more configs than required. Removed configs with smallest crowding distance.")

        return selected_configs

    def _get_next_order_seed(self) -> int | None:
        """Next instances shuffle seed to use."""
        # Here we have the option to shuffle the trials when specified by the user
        if self._instance_seed_order == "shuffle":
            seed = self._rng.randint(0, MAXINT)
        elif self._instance_seed_order == "shuffle_once":
            seed = 0
        else:
            seed = None

        return seed

    def _get_next_bracket(self) -> int:
        """Successive Halving only uses one bracket. Therefore, we always return 0 here."""
        return 0
